// -*- C++ -*-

// DataPipe.h
//
// Classes for sources and sinks of data
// - schedulable using Scheduler
// - transferring data through CircularBuffers
// for Audio Pattern Recognition Language (aprl)
// dpwe@media.mit.edu 1995feb02
//

#ifndef _DATAPIPE_H
#define _DATAPIPE_H

#include "CircularBuffer.h"
#include "Scheduler.h"

#include <math.h>	// for ceil()

extern "C" void abort(void);	// from stdlib.h 

class DataSink;		// forward declaration allows list class definition

// MAPpable function type applied to a process, taking pointer to private data
typedef void DataSinkFn(DataSink *d, void *dpfdata);

class DataSinkNode {	// Node for the list of DataSinks we will define
public:
    DataSinkNode *next;
    DataSink 	 *data;

    DataSinkNode(DataSink *dp, DataSinkNode *nx = NULL) 
        { data = dp; next = nx; }
    ~DataSinkNode(void)	
        { };
};

class DataSinkList  {
private:
    DataSinkNode *head;
public:
    DataSinkList(void)	
        { head = 0; }
    ~DataSinkList(void);
    void add(DataSink *proc);		// add to top of list
    int  remove(DataSink *proc);	// find node matching proc, remove
    void apply(DataSinkFn *dpf, void *dpfdata);     // map dpf to each node
    double earliestSequence(void);	// find the smallest sq val in list
    DataSinkList *Duplicate(void);	// make a copy of top-leve
};

class DataSource : public virtual SchedProc, public CircularBuffer {
    // Schedulable function produces data output which is stored 
    // in the circular buffer from which it inherits
private:
    DataSinkList *outputs;	// list of all the registered users of output
    double sampleRate;		// sampling rate of output data
protected:
public:
    DataSource(double srate=0, int bufsz = 0) 
        : CircularBuffer(bufsz), sampleRate(srate)
	{ outputs = new DataSinkList(); };
    DataSource(const DataSource &dsc) : CircularBuffer(dsc.Size())
        { sampleRate = dsc.sampleRate;
	  fprintf(stderr, "DSrc0x%lx: CopyCon: sr=%.0f\n", this, sampleRate);
          outputs = 0; };
    virtual ~DataSource(void);	// make sure all Sinks know we're deserting them
    // Accessors
    double SampleRate(void) const	{ return sampleRate; }
    double CountTime(void) const	{ return Count()/sampleRate; }
    // Methods
    void RegisterSink(DataSink *sink)	// Add a particular sink to our
    	{ outputs->add(sink); };	//   list of recipients
    void UnregisterSink(DataSink *sink) // remove from sink list
	{ assert(outputs->remove(sink)); };
    double EarliestOutputSequence(void) // min of seqs of outputs
        { return outputs->earliestSequence(); }
    // inherits circbuffer's GetHeadPtr, Count
    // still pure virtual because Update() method not defined.
    void SetSampleRate(double srate)	{ sampleRate = srate; 
				          /* notify outputs.. */ }
    virtual int ThereIsMore(void)	// allows test for end-of-data
       { return 1; }	// by default, always true
    void EnsureSeqAbove(double ipseq)   // make our sequence more than ipsq
	// but make sure we stay below the sq of our outputs
	{ double minoutseq = EarliestOutputSequence();
	  double myseq = Sequence();
	  if(ipseq > minoutseq)  {
	      fprintf(stderr, "DSrc:EnsSqAbv:op minoseq %.1f > ipseq %.1f\n",
		      minoutseq, ipseq);
	      abort();
	  }
	  if(!(ipseq < myseq && myseq < minoutseq))
	      SetSequence(0.5*(ipseq+minoutseq));
        }
};


class DataSink : virtual public SchedProc  {
    // schedulable objects accepting data from DataSources
private:
    DataSource *input;	// where our data is to be obtained
// protected:
//    int inputCount;	// we've processed input up to this count.
public:
    DataSink(DataSource *in=0)  { input = 0; SetInput(in);
    				  /* inputCount = 0; */ }
    virtual ~DataSink(void)	{ if(input) input->UnregisterSink(this); }
    
    virtual void SetInput(DataSource *in=0)	
	{ if(input) input->UnregisterSink(this);
	  input = in;
	  if(input) {
	      double ipseq = input->Sequence();
	      if(Sequence() <= ipseq)
		  SetSequence(1+ipseq);
	      input->RegisterSink(this); } };
    DataSource *Input(void) const { return input; }
    int ThereIsMore(void)	// pass query back to input
    { int rc = 0;
      if(!IsScheduled())	fprintf(stderr,"DataSink: not scheduled\n");
      else if(Input()==0)	fprintf(stderr, "DataSink: no input\n");
      else			rc = Input()->ThereIsMore(); 
      return rc; }
};

class NullDataSink : public DataSink {	// a non-pure-virtual instance I need
public:
    NullDataSink(DataSource *in=0)	:	DataSink(in)	{ };
    virtual ~NullDataSink(void)					{ };
    int Update(double /* time */)				{ return 0; }
};

class DataPipe : public DataSource, public DataSink  {  // input/output class
private:
    double ioRatio;	// number of input samples used for each op sample
    void updateSrate(void)
        { if(Input())  {  SetSampleRate(Input()->SampleRate()/ioRatio); } };
public:
    DataPipe(DataSource *in = 0, double iorat = 1.0) : 
    	DataSink(), ioRatio(iorat), DataSource()
	{ if(in) SetInput(in);
	  updateSrate(); }
    virtual ~DataPipe(void)	{ };
    void SetIoRatio(double iorat)
        { ioRatio = iorat; updateSrate(); }
    void SetInput(DataSource *in=0)	
        { DataSink::SetInput(in);
	  if(Input()) {
	      if(Size()==0)   // need to set up size of our circ buffer
		  Resize((int)ceil(ioRatio * Input()->Size()));
	      updateSrate();
	      double ipseq = Input()->Sequence();
	      EnsureSeqAbove(ipseq);
	  } };
    virtual int ThereIsMore(void)
        { return DataSink::ThereIsMore(); }
};

class DataPipeProcess : public DataPipe {  // IO with a simple function
protected:
    DATAprocess   *procFn;
public:
    DataPipeProcess(DataSource *in = 0, 
		    double iorat = 1.0, DATAprocess *pfn=0)
	: procFn(pfn), DataPipe(in, iorat)
        { };
    virtual ~DataPipeProcess(void)
        { };
    int Update(double /* time */)
	{    /* actually invoke the process on the data. <time> ignored? */
	    int done=0 /* , last */;
	    
//	    done = inputCount;	// last input point we have processed
//	    last = Input()->Count();  // input points now available
	    if(ThereIsMore())  {
		PushHead();	// subsequent PopTail()s on us will go here
		assert(Input()->PopTail()>=0);  // reset to read from input
		done = CircBufInputOutput((CircularBuffer *)Input(),
					  (CircularBuffer *)this, procFn);
	    }
	    return done;
	};
};

class DataPipeSimple : public DataPipe {  // IO with a simple function
protected:
    DATAsimpleFn   *procFn;
public:
    DataPipeSimple(DataSource *in = 0, DATAsimpleFn *pfn=0)
	: procFn(pfn), DataPipe(in, 1.0)
        { };
    virtual ~DataPipeSimple(void)
        { };
    int Update(double /* time */)
	{    /* actually invoke the process on the data. <time> ignored? */
	    int done=0 /* , last */;
	    
	    if(ThereIsMore())  {
		PushHead();	// subsequent PopTail()s on us will go here
		assert(Input()->PopTail()>=0);  // reset to read from input
		done = CircBufSimpleIO((CircularBuffer *)Input(),
				      (CircularBuffer *)this, procFn);
	    }
	    return done;
	};
};

#endif /* _DATAPIPE_H */
