//
// Filter.cc
// 
// Implementation of digital filter class
// dpwe@interval.com 1994aug11
//

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <stdio.h>

#include "Filter.h"

#include "dpwedefs.h"

// ------------------------ FilterDef -------------------------------
//  *defines* a filter through a and b vectors - but no state.  
//  Thus several Filter objects can share the same FilterDef (to implement
//  the same filter) but process different channels, having different 
//  state

/*
ostream& operator<< (ostream& f, const FilterDef& a)
{   // write A descriptor to output
    int i;
    char str[32];

    sprintf(str, "#FD<0x%lx,%d>", &a, a.order());
    f << str;
    return f;
}
*/

// ------------------------ Filter ----------------------------------
//  contains a pointer to a filterDef (allowing several filters to use 
//  the same coefficients) and a state vector, has methods to apply 
//  the filter to data.

void Filter::_makeFdef(doubleVec &as, doubleVec &bs)
{   /* create a new fdef based on the sets of coefs, and point ourselves 
       to it */
    fdef  = new FilterDef(as, bs);
    ownfd = 1;				// remember to delete FD
    wi.resize(0);  wi.resize(fdef->order());	// clears too
    wibase = 0;	// next entry of wi to be written
    fdef->as.resize(fdef->order());	// make sure they're all same length,
    fdef->bs.resize(fdef->order());	// zero padded.  allows fast loop
}

void Filter::_changeFdef(doubleVec &as, doubleVec &bs)
{   /* dynamically (!?) modify the filter definition */
    if(ownfd)	delete fdef;
    fdef  = new FilterDef(as, bs);
    assert(fdef->order() == wi.size());	// constraint on changing
    ownfd = 1;				// remember to delete FD
    fdef->as.resize(fdef->order());	// make sure they're all same length,
    fdef->bs.resize(fdef->order());	// zero padded.  allows fast loop
    // leave the current state as it is
}

Filter::Filter(doubleVec &as, doubleVec &bs) 
{   /* Make a new filter with the specified coefficients
       Coefficient arrays are copied, remain property of caller */
    _makeFdef(as, bs);
}

Filter::Filter(const Filter& filt)
{   /* copy constructor */
    /* copying means that we want the same fdef.
       But if the thing we're copying from *owns* the fdef, then maybe 
       we should make a copy of the fdef too, that we can own.  Then 
       if the thing we're copying is a temporary that's destroyed, we're 
       OK.  Plus we're really a copy */
    if(filt.ownfd)  {
	ownfd = 1;
	fdef = new FilterDef(*filt.fdef);
    } else {
	ownfd = 0;
	fdef = filt.fdef;
    }
    wi.resize(0); wi.resize(fdef->order()); wibase = 0;
}

Filter& Filter::operator = (const Filter& filt)
{   /* reassignment to this object */
    if(this != &filt)  {
	// destructor part
	if(ownfd)  {
	    delete fdef;
	}
	// copy constructor part
	if(filt.ownfd)  {
	    ownfd = 1;
	    fdef = new FilterDef(*filt.fdef);
	} else {
	    ownfd = 0;
	    fdef = filt.fdef;
	}	
	wi.resize(0);  wi.resize(fdef->order());
    }
    return *this;
}

Filter::~Filter()
{   /* destructor */
    if(ownfd)  {	// if we created the fdef we point to, destroy it.
	delete fdef;
    }
    // destructor for wi is called automatically
}

void Filter::oldrun(float *in, float *out, int points)
{   /* apply the filter to some new input */
    int i,j;
    double w;
    
//    DBGFPRINTF((stderr, "filter::run in %lx, out %lx, pts %ld\n", 
//		in, out, points));
    for(i = 0; i<points; ++i)  {
      for(j=wi.size()-1; j>0; --j) {
        wi[j] = wi[j-1];
      }
      w = fdef->as[0] * *in++;	// as[0] arbitrarily scales entire filter
      for(j=1; j<fdef->as.size(); ++j)  {
        w -= wi[j] * fdef->as[j];  // minus because of definition
      }
      wi[0] = w;
      // now w becomes new output
      w *= fdef->bs[0];
      for(j=1; j<fdef->bs.size(); ++j)  {
        w += fdef->bs[j]*wi[j];
      }
      *out++ = w;
    }
}

void Filter::run53(float *in, float *out, int points)
{   /* Special optimized run version for filters with 4 poles and 2 zeros
       (our favorite bpf_2ndo) */
    int i;
    double w0, w1, w2, w3, w4, a1, a2, a3, a4, b0, b1, b2;
    
    w0 = wi.elem(0);
    w1 = wi.elem(1);
    w2 = wi.elem(2);
    w3 = wi.elem(3);
    a1 = fdef->as.elem(1);
    a2 = fdef->as.elem(2);
    a3 = fdef->as.elem(3);
    a4 = fdef->as.elem(4);
    b0 = fdef->as.elem(0) * fdef->bs.elem(0);
    b1 = fdef->as.elem(0) * fdef->bs.elem(1);
    b2 = fdef->as.elem(0) * fdef->bs.elem(2);
    for(i = 0; i<points; ++i)  {
	w4 = w3; 	w3 = w2; 	w2 = w1;	w1 = w0;
	w0 = *in++ - a1*w1 - a2*w2 - a3*w3 -a4*w4;
	*out++ = b0*w0 + b1*w1 + b2*w2;
    }
    wi.elem(0) = w0;
    wi.elem(1) = w1;
    wi.elem(2) = w2;
    wi.elem(3) = w3;
}
    
void Filter::run(float *in, float *out, int points)
{   /* apply the filter to some new input.  time testfilt = 0.825u */
    int i, orderm1 = wi.size()-1;
    double w0,x0,wn;
    double *wp, *ap, *bp, *at, *bt, *wb;
    int wib = wibase;
    int firstpass, secondpass;
/* optimization tests:  
   Running bpf over timetest.aif: 
       run:	2.30s	run55:	0.40 (5.8x)
   running hpf over timetest.aif:
       run:	1.58s	run33:	0.39 (4.0x)

*/
    if(wi.size()==3)  {
	run33(in, out, points);
	return;
    } else if(wi.size()==5){
	if(fdef->bs.elem(3)==0 && fdef->bs.elem(4)==0)  {
	    run53(in,out,points);	/* common case - superfastest */
	} else {
	    run55(in, out, points);
	}
	return;
    }

    at = &(fdef->as.elem(orderm1));
    bt = &(fdef->bs.elem(orderm1));
    wb = &(wi.elem(0));
    wp = wb+wib;

    for(i = 0; i<points; ++i)  {
	ap = at;
	bp = bt;
	w0 = *in++;	
	x0 = 0;
	wp = wb+wib;
	firstpass = orderm1-wib;
	secondpass = orderm1-firstpass;
	while(firstpass--)  {
	    wn = *wp++;
	    w0 -= wn * *ap--;
	    x0 += wn * *bp--;
	}
	// by def, wp now == wb+orderm1
	wp = wb;
	while(secondpass--)  {
	    wn = *wp++;
	    w0 -= wn * *ap--;
	    x0 += wn * *bp--;
	}
	// now wp == wb+wib again
	*wp = w0;
	*out++ = *ap-- * (x0 + *bp-- * w0);
	// as[0] arbitrarily scales output

	wib = (wib+1)%orderm1;
    }
    wibase = wib;
}
    
void Filter::run33(float *in, float *out, int points)
{   /* Special optimized run version for 3rd order filters */
    int i;
    double w0, w1, w2, a1, a2, b0, b1, b2;
    
    w0 = wi.elem(0);
    w1 = wi.elem(1);
    w2 = wi.elem(2);
    a1 = fdef->as.elem(1);
    a2 = fdef->as.elem(2);
    b0 = fdef->as.elem(0) * fdef->bs.elem(0);
    b1 = fdef->as.elem(0) * fdef->bs.elem(1);
    b2 = fdef->as.elem(0) * fdef->bs.elem(2);
    for(i = 0; i<points; ++i)  {
	w2 = w1;	w1 = w0;
	w0 = *in++ - a1*w1 - a2*w2;
	*out++ = b0*w0 + b1*w1 + b2*w2;
    }
    wi.elem(0) = w0;
    wi.elem(1) = w1;
    wi.elem(2) = w2;
}
    
void Filter::run55(float *in, float *out, int points)
{   /* Special optimized run version for 5th order filters time=656 =+25%  */
    int i;
    double w0, w1, w2, w3, w4, a1, a2, a3, a4, b0, b1, b2, b3, b4;
    
    w0 = wi.elem(0);
    w1 = wi.elem(1);
    w2 = wi.elem(2);
    w3 = wi.elem(3);
    w4 = wi.elem(4);
    a1 = fdef->as.elem(1);
    a2 = fdef->as.elem(2);
    a3 = fdef->as.elem(3);
    a4 = fdef->as.elem(4);
    b0 = fdef->as.elem(0) * fdef->bs.elem(0);
    b1 = fdef->as.elem(0) * fdef->bs.elem(1);
    b2 = fdef->as.elem(0) * fdef->bs.elem(2);
    b3 = fdef->as.elem(0) * fdef->bs.elem(3);
    b4 = fdef->as.elem(0) * fdef->bs.elem(4);
    for(i = 0; i<points; ++i)  {
	w4 = w3; 	w3 = w2; 	w2 = w1;	w1 = w0;
	w0 = *in++ - a1*w1 - a2*w2 - a3*w3 -a4*w4;
	*out++ = b0*w0 + b1*w1 + b2*w2 + b3*w3 + b4*w4;
    }
    wi.elem(0) = w0;
    wi.elem(1) = w1;
    wi.elem(2) = w2;
    wi.elem(3) = w3;
    wi.elem(4) = w4;
}
    
void Filter::print(char *title)
{  
    if(title == 0)	title = "";
    printf("Filter %s: 0x%lx, order %d:\n", title, this, fdef->order());
    int i;
    printf("As(%d):", fdef->as.size());
    for(i=0; i<fdef->as.size(); ++i)
	printf("%.3f ", fdef->as[i]);
    printf("\nBs(%d):", fdef->bs.size());
    for(i=0; i<fdef->bs.size(); ++i)
	printf("%.3f ", fdef->bs[i]);
    printf("\n");
}
    
static void convolve(doubleVec & rslt, doubleVec & with)
{
    int i,j,m;
    int nr = rslt.size(), nw = with.size();
    double s;
	
    // does not extend result - extra bits fall off end
    for(j = nr-1; j >=0; --j)  {
	s = 0;
	m = min(j+1, nw);
	for(i = 0; i<m; ++i)  {
	    s += (rslt)[j-i] * (with)[i];
	}
	(rslt)[j] = s;
    }
}

// class BandpassFilter : public Filter  
// special case : filter with BP descr.

#ifndef M_PI
#define M_PI 3.1415926535
#endif				/* M_PI */

#ifdef USE_COMPLEX
#ifdef __GNUC__
#include <Complex.h>
#else /* POCC */
#include <complex.h>
#define Complex complex
#endif /* GNUG/Cfront */
#endif /* USE_COMPLEX */

void BandpassFilter::SecondOrderFilter(double ctrFreq, double Qfactor, 
				       int repeat, 
				       doubleVec &ai, doubleVec &bi)
{   /* construct a bandpass filter with the given characteristics */
    double cfRad = M_PI*2.0*ctrFreq/sampRate;
    double desQ = Qfactor/double(repeat);
    double R = 1.0 - (cfRad/(2*desQ));
    double G = 0.447; 	// this value preserves RMS for 40 bins at 6 bpo
                        // from 100 Hz on a 22050 kHz sampled system
                        // when you just add all the channels together
//pow((double)(1.0-R), (double)repeat)*(ctrFreq/Qfactor);
    int i;
    doubleVec basicA(3);
    
//    fprintf(stderr, "BPF:2O: cf %.0f sr %.0f Q %.1f rpt %d\n", 
//	    ctrFreq, sampRate, Qfactor, repeat);

    basicA[0] = 1.0;  basicA[1] = -2.0*R*cos(cfRad); basicA[2] = R*R;
    int order = 1+2*repeat;
//  doubleVec ai(order, 0.0);
    ai.resize(order);
    ai[0] = 1.0;
    for(i = 0; i<repeat; ++i)
	convolve(ai, basicA);
    // put zeros at dc and fnyq
//  doubleVec bi(3);
    bi.resize(3);
    bi[0] =  G; bi[1] = 0; bi[2] = -G;
    // fix the gain
#ifdef USE_COMPLEX
    Complex p1(R*cos(cfRad), R*sin(cfRad));
    Complex p2(R*cos(cfRad), -R*sin(cfRad));
    Complex z(cos(cfRad), sin(cfRad));
//    Complex mag = (1-z)*(-1-z)/((p1-z)*(p2-z));
    Complex mag = 1.0/((p1-z)*(p2-z));
    // poles are repeated, so factor out the power of their abs()
    // then divide by the gain due to the zeros just once.
    ai[0] = pow(1.0/abs(mag), (double)repeat)/abs((1-z)*(-1-z));
#else /* !USE_COMPLEX */
    double mag = 1.0/((1.0 - R) * sqrt(1+R*R-2*R*cos(2*cfRad)));
    ai[0] = pow(1.0/mag, (double)repeat) / (4*sin(cfRad/2)*sin((M_PI-cfRad)/2));
    // I derived these geometrically, looking at the Zplane the compex
    // was trying to solve
    // I checked (in matlab) that this matches the complex math for
    // R = 0.7, cfRad = 0.1 - 1997jul02 dpwe@icsi.berkeley.edu
#endif /* USE_COMPLEX */

//    ai[0] = (sqrt(1000.0/ctrFreq));		// !!!!!!!!!!!!!!!!!! hack
    // _makeFdef(ai, bi);
    // set up description - cf, Q, tdel
    ctrFrq = ctrFreq;
    Qfctr  = Qfactor;
    g      = 1.0;		// gain (at ctr frq) - not sure!!
    smpDly = Qfctr/(2*repeat)*sampRate/ctrFreq; // proportional to 1/bandwidth
}

void BandpassFilter::PattersonFilter(double cf, doubleVec &a, doubleVec &b)
{   /* construct a Patterson-Slaney Gammatone filter at this cf */
    /* design a patterson-slaney bandpass filter whose bandwidth matches 
       the frequency specified.  Return its bandwidth and group delay. */
    /* After Malcolm Slaney's Matlab fn.  */
    double earQ = 9.26449; /* Glasberg/Moore value */
    double minBW = 24.7;
    int    order = 4;
    double pi    = M_PI;
    double ERB   = pow(pow(cf/earQ, double(order))
		       +pow(minBW, double(order)), 1.0/order);
    double B     = 1.019*2*pi*ERB;
    double T     = 1.0/sampRate;
    double g1fR  = -2*cos(4*cf*pi*T) + 2*exp(-(B*T))*cos(2*cf*pi*T);
    double g1fI  = -2*sin(4*cf*pi*T) + 2*exp(-(B*T))*sin(2*cf*pi*T);
    double c2fpt = cos(2*cf*pi*T);
    double s2fpt = sin(2*cf*pi*T);
    double r3p232= sqrt(3+pow(2,1.5));
    double r3m232= sqrt(3-pow(2,1.5));
    double re1, im1, re2, im2, re3, im3, denoR, denoI, gain;
//    doubleVec a(9), b(5);
    a.resize(9);	b.resize(5);
    int    /* len = 9, */ n = 4;
    
    re1 = g1fR*T*(c2fpt - r3m232*s2fpt);  im1 = g1fI*T*(c2fpt - r3m232*s2fpt);
    re2 = g1fR*T*(c2fpt + r3m232*s2fpt);  im2 = g1fI*T*(c2fpt + r3m232*s2fpt);
    re3 = re1*re2 - im1*im2;	
    im3 = re1*im2 + re2*im1;
    re2 = g1fR*T*(c2fpt - r3p232*s2fpt);  im2 = g1fI*T*(c2fpt - r3p232*s2fpt);
    re1 = re3*re2 - im3*im2;	
    im1 = re3*im2 + re2*im3;
    re2 = g1fR*T*(c2fpt + r3p232*s2fpt);  im2 = g1fI*T*(c2fpt + r3p232*s2fpt);
    re3 = re1*re2 - im1*im2;	
    im3 = re1*im2 + re2*im1;
    
    denoR = -2*exp(-2*B*T) - 2*cos(4*cf*pi*T) + 2*exp(-B*T)*(1+cos(4*cf*pi*T));
    denoI =                 -2*sin(4*cf*pi*T) + 2*exp(-B*T) *  sin(4*cf*pi*T);
    
    gain = sqrt(re3*re3 + im3*im3) / pow(sqrt(denoR*denoR + denoI*denoI),4.0);
    
    /* original matlab:
       gain = abs( (-2*exp(4*i*cf*pi*T)*T + 2*exp(-(B*T) + 2*i*cf*pi*T)
       *T*(cos(2*cf*pi*T) - sqrt(3 - 2^(3/2))*sin(2*cf*pi*T)))
       *(-2*exp(4*i*cf*pi*T)*T + 2*exp(-(B*T) + 2*i*cf*pi*T)
       *T*(cos(2*cf*pi*T) + sqrt(3 - 2^(3/2))*sin(2*cf*pi*T)))
       *(-2*exp(4*i*cf*pi*T)*T + 2*exp(-(B*T) + 2*i*cf*pi*T)
       *T*(cos(2*cf*pi*T) - sqrt(3 + 2^(3/2))*sin(2*cf*pi*T)))
       *(-2*exp(4*i*cf*pi*T)*T + 2*exp(-(B*T) + 2*i*cf*pi*T)
       *T*(cos(2*cf*pi*T) + sqrt(3 + 2^(3/2))*sin(2*cf*pi*T)))
       /(-2 / exp(2*B*T) - 2*exp(4*i*cf*pi*T) 
       + 2*(1 + exp(4*i*cf*pi*T))/exp(B*T))
       ^4);
       */
    b[0] =    pow(T, 4.0)/gain;
    b[1] = -4*pow(T, 4.0)*cos(2*cf*pi*T)/exp(B*T)/gain;
    b[2] =  6*pow(T, 4.0)*cos(4*cf*pi*T)/exp(2*B*T)/gain;
    b[3] = -4*pow(T, 4.0)*cos(6*cf*pi*T)/exp(3*B*T)/gain;
    b[4] =    pow(T, 4.0)*cos(8*cf*pi*T)/exp(4*B*T)/gain;
    
    a[0] =                                              1.0;
    a[1] =                            -8*cos(2*cf*pi*T)/exp(B*T);
    a[2] =                     4*(4 + 3*cos(4*cf*pi*T))/exp(2*B*T);
    a[3] =       -8*(6*cos(2*cf*pi*T) + cos(6*cf*pi*T))/exp(3*B*T);
    a[4] =  2*(18 + 16*cos(4*cf*pi*T) + cos(8*cf*pi*T))/exp(4*B*T);
    a[5] =       -8*(6*cos(2*cf*pi*T) + cos(6*cf*pi*T))/exp(5*B*T);
    a[6] =                     4*(4 + 3*cos(4*cf*pi*T))/exp(6*B*T);
    a[7] =                            -8*cos(2*cf*pi*T)/exp(7*B*T);
    a[8] =                                            1/exp(8*B*T);
    
    // _makeFdef(a, b);
    // set up description - cf, Q, tdel
    ctrFrq = cf;
    Qfctr  = ERB/ctrFrq;
    g      = 1.0;		// gain (at ctr frq) - not sure!!
    smpDly = sampRate*(n-1)/B;	// proportional to bandwidth
}

BandpassFilter::BandpassFilter(double sRate, double ctrFreq, double Qfactor, 
			       int typ)
{   /* construct a bandpass filter with the given characteristics */
    sampRate = sRate;
    type     = typ;
    doubleVec a(0), b(0);
    switch(type)  {
    case BPF_2NDO:
	SecondOrderFilter(ctrFreq, Qfactor, /* repeated */ 2, a, b);
	_makeFdef(a,b);
	break;
    case BPF_PATT:
	PattersonFilter(ctrFreq, a, b);
	_makeFdef(a,b);
	break;
    default:
	fprintf(stderr, "BandpassFilter: unknown type %d\n", type);
	break;
    }
}

void BandpassFilter::ChangeFilter(double ctrFreq, double Qfactor)
{   /* revise chrs of bandpass filter */
    doubleVec a(0), b(0);
    switch(type)  {
    case BPF_2NDO:
	SecondOrderFilter(ctrFreq, Qfactor, /* repeated */ 2, a, b);
	_changeFdef(a,b);
	break;
    case BPF_PATT:
	PattersonFilter(ctrFreq, a, b);
	_changeFdef(a,b);
	break;
    default:
	fprintf(stderr, "BandpassFilter::Change: unknown type %d\n", type);
	break;
    }
}

static void LPFdesign(double sRate, double cutFreq, double /* Qfactor */ , 
		      int type, doubleVec &ai, doubleVec &bi)
{   /* construct a lowpass filter with the given characteristics */
    double alpha;
    
    switch(type)  {
    case LPF_1STO:
	ai.resize(2);
	bi.resize(1);
	alpha = exp(-(1.0/sRate)/(1.0/cutFreq));
	ai[0] = 1; ai[1] = -alpha;
	bi[0] = 1.0-alpha;
	break;
    case LPF_2NDO:
	ai.resize(3);
	bi.resize(1);
	alpha = exp(-(1.0/sRate)/(1.0/cutFreq));
	ai[0] = 1; ai[1] = -2*alpha; ai[2] = alpha*alpha;
	bi[0] = (1.0-alpha)*(1.0-alpha);
	break;
    default:
	fprintf(stderr, "LowpassFilter: unknown type %d\n", type);
	break;
    }
}

LowpassFilter::LowpassFilter(double sRate, double cutFreq, double Qfactor, 
			     int typ)
{   /* construct a lowpass filter with the given characteristics */
    doubleVec ai, bi;
    
    sampRate = sRate;
    type     = typ;
    cutoffFrq= cutFreq;
    Qfctr    = Qfactor;

    LPFdesign(sampRate, cutoffFrq, Qfctr, type, ai, bi);
    _makeFdef(ai, bi);
}

void LowpassFilter::ChangeFilter(double cutFreq, double Qfactor)
{   /* revise chrs of lowpass filter */
    doubleVec a(0), b(0);

    cutoffFrq= cutFreq;
    Qfctr    = Qfactor;
    LPFdesign(sampRate, cutoffFrq, Qfctr, type, a, b);
    _changeFdef(a,b);
}


static void HPFdesign(double sRate, double cutFreq, double /* Qfactor */ , 
		      int type, doubleVec &ai, doubleVec &bi)
{   /* construct a highpass filter with the given characteristics */
    double alpha, cfrad;
    
    switch(type)  {
    case HPF_1STO:	// a zero at dc and a pole a little inside
	ai.resize(2);
	bi.resize(2);
	alpha = exp(-(1.0/sRate)/(1.0/cutFreq));
	ai[0] = 1.0;       ai[1] = -alpha;
	bi[0] = 1.0; bi[1] = -1.0;
	break;
    case HPF_2NDO:	// a zero at dc and conjugate poles
	ai.resize(3);
	bi.resize(3);
	cfrad = M_PI*2.0*cutFreq/sRate;
	alpha = exp(-(sin(cfrad)/sin((M_PI-cfrad)/2)));
        ai[0] = 1; 
	ai[1] = -2*alpha*cos(cfrad);
	ai[2] = alpha*alpha;
	bi[0] = 1.0;	bi[1] = -2.0;	bi[2] = 1.0;
	break;
    default:
	fprintf(stderr, "HighpassFilter: unknown type %d\n", type);
	break;
    }
}

HighpassFilter::HighpassFilter(double sRate, double cutFreq, double Qfactor, 
			     int typ)
{   /* construct a highpass filter with the given characteristics */
    doubleVec ai, bi;
    
    sampRate = sRate;
    type     = typ;
    cutoffFrq= cutFreq;
    Qfctr    = Qfactor;

    HPFdesign(sampRate, cutoffFrq, Qfctr, type, ai, bi);
    _makeFdef(ai, bi);
}

void HighpassFilter::ChangeFilter(double cutFreq, double Qfactor)
{   /* revise chrs of highpass filter */
    doubleVec a(0), b(0);

    cutoffFrq= cutFreq;
    Qfctr    = Qfactor;
    HPFdesign(sampRate, cutoffFrq, Qfctr, type, a, b);
    _changeFdef(a,b);
}


#ifdef MAIN

#include <stdio.h>

void print(FILE *stream, float *buf, int len, char *title=0)
{   /* print out some float data, suitable for xgraph */
    int i;
    
    fprintf(stream, "\n");
    if(title)
	fprintf(stream, "\"%s\"\n", title);
    for(i=0; i<len; ++i)  {
	fprintf(stream, "%d %.3f\n", i, buf[i]);
    }
}

main(int, char **)
{
    int dsize = 100;
    float *ibuf = new float[dsize];
    float *obuf = new float[dsize];
    int i;
    
    for(i=0; i<dsize; ++i)
	ibuf[i] = 0.0;
    
    ibuf[5] = 1.0;
    
    //    BandpassFilter bpf(10000, 1000, 8, BPF_PATT);
    
    LowpassFilter bpf(10000, 1000, 8, LPF_2NDO);
    
    print(stdout, ibuf, dsize, "impulse");
    
    bpf.run(ibuf, obuf, dsize);
    
    print(stdout, obuf, dsize, "response - 1/10");
    
    bpf = LowpassFilter(10000, 500, 8, LPF_2NDO);
    
    bpf.run(ibuf, obuf, dsize);
    
    print(stdout, obuf, dsize, "response - 1/20");
    
    delete [] ibuf;
    delete [] obuf;
}

#endif				/* MAIN */
