// -*- C++ -*-
// 
// Filter.h
// 
// Regular digital filter class
// dpwe@interval.com
// 1994aug11
// 

#ifndef _FILTER_H
#define _FILTER_H

#include "double.Vec.h"

class FilterDef  {  // define a digital filter by as and bs
public:
    doubleVec	as;
    doubleVec	bs;

    FilterDef() 			: as(0), bs(0) 		
                                     { /* cerr << "FD()"<< *this <<"\n"; */ };
    FilterDef(const doubleVec& a, const doubleVec& b) 	: as(a), bs(b)
                                     { /* cerr << "FD(ab)"<< *this <<"\n";*/ };
    FilterDef(const FilterDef& f) 		: as(f.as), bs(f.bs)
                                     { /* cerr << "FD(fd)"<< *this <<"\n";*/ };
    ~FilterDef()
                                     { /* cerr << "~FD()"<< *this <<"\n"; */ };
    int order(void) const {return(as.size()>bs.size())? as.size():bs.size(); };

//    friend ostream& operator<< (ostream& f, const FilterDef& a);
};

class Filter {	// A regular digital filter with state
private:
  FilterDef *fdef;
  int ownfd;		// flag as whether we created (& should destroy) *fdef
  doubleVec wi;
  int wibase;
  /* special case filter routines */
  void run33(float *in, float *out, int points);
  void run55(float *in, float *out, int points);
  void run53(float *in, float *out, int points);
protected:
  void _makeFdef(doubleVec &a, doubleVec &b);// copy coefs, make private fdef
  void _changeFdef(doubleVec &a, doubleVec &b);  // replace existing fdef
public:
  Filter() : fdef(0), ownfd(0), wi(0), wibase(0)			{};
  Filter(FilterDef& fd) { _makeFdef(fd.as, fd.bs); }
  Filter(doubleVec &a, doubleVec &b);
  Filter(const Filter& filt);
  Filter& operator = (const Filter& filt);
  ~Filter(void);
  void run(float *in, float *out, int points);
  void oldrun(float *in, float *out, int points);
  void print(char *title = 0);
  int  order(void) const  { return fdef->order(); }
};

enum {	// types for BandpassFilter(type)
    BPF_2NDO = 1,	// regular 2nd order bandpass filter
    BPF_PATT		// Patterson-Slaney gammatone filter
};

class BandpassFilter : public Filter  {   // special case : filt with BP descr.
private:
    double	sampRate;
    double	ctrFrq;
    double	Qfctr;
    double	g;
    double	smpDly;
    int		type;

    void 	PattersonFilter(double cf, doubleVec &a, doubleVec &b);
    void 	SecondOrderFilter(double ctrFreq, 
				  double Qfactor, int repeat, 
				  doubleVec &a, doubleVec &b);
public:
    BandpassFilter(double srate, double ctrFreq, double Qfactor, int type);
    BandpassFilter(const BandpassFilter& f) : Filter(f)
       {  sampRate=f.sampRate; ctrFrq=f.ctrFrq; Qfctr=f.Qfctr;
	  g = f.g; smpDly = f.smpDly; type = f.type; }
    void 	ChangeFilter(double ctrFreq, double Qfactor);

    double	sampleRate(void)  const { return sampRate; };
    double	centerFreq(void)  const { return ctrFrq;   };
    double	Qfactor(void)     const { return Qfctr;    };
    double	gain(void)	  const { return g;	      };
    double	sampleDelay(void) const { return smpDly;   };
};

enum {	// types for lowpass filter
    LPF_1STO = 1,
    LPF_2NDO 
};

class LowpassFilter : public Filter  {  // special case for lowpass (smooth)
private:
    double	sampRate;
    double	cutoffFrq;
    double	Qfctr;
    int 	type;
public:
    LowpassFilter(double srate, double cofrq, double Qfactor, int type);
    void 	ChangeFilter(double coFreq, double Qfactor);

    double	sampleRate(void)   { return sampRate; };
    double	cutoffFreq(void)   { return cutoffFrq;   };
    double	Qfactor(void)      { return Qfctr;    };
};
    
enum {	// types for highpass filter
    HPF_1STO = 1,
    HPF_2NDO 
};

class HighpassFilter : public Filter  {  // special case for highpass (blockdc)
private:
    double	sampRate;
    double	cutoffFrq;
    double	Qfctr;
    int 	type;
public:
    HighpassFilter(double srate, double cofrq, double Qfactor, int type);
    void 	ChangeFilter(double coFreq, double Qfactor);

    double	sampleRate(void)   { return sampRate; };
    double	cutoffFreq(void)   { return cutoffFrq;   };
    double	Qfactor(void)      { return Qfctr;    };
};
    
#endif /* _FILTER_H */
