/*
 * Copyright 1994 Massachusetts Institute of Technology
 *
 * This software may be used for research purposes only. By prior
 * agreement with Lexicon, Inc. no part of this software may be
 * sold for profit or used in any commercial product, and the details
 * of the algorithms used may not be published or publicly presented.
 */
/*
 * Pitch tracker.
 *
 * Period detector pitch tracker. This works using a single
 * period detector that has both tracking and acquire modes. Zero crossing
 * events are determined in the module "Xing.c", and this module determines
 * periodicities in the stream of events. The output of this module is
 * the current period estimate, returned by calling the function
 * ptrack_get_period(). Note tracking state is communicated by the
 * callback function note_on().
 *
 * The pitch tracker is surrounded by logic that generates note_on
 * and note_off events. This logic allows for two important features:
 *    - Programmable number of periods tracked before pitch is
 *	considered valid.
 *    - Programmable number of dropout periods tolerated without
 *	generating a new note_on event. A dropout is when the pitch
 *	tracker loses track of a pitch (i.e. goes back to acquire
 *	mode), and then re-acquires the same pitch.
 *
 * Bill Gardner, January 1994.
 */
#include    <stdio.h>
#include    <math.h>
#include    <stdlib.h>
#include    <string.h>

extern "C" {
#ifdef BILLG
#include    "misc.h"
#include    "userface.h"
#else 
#define TRUE (1)
#define FALSE (0)  
#endif
};
#include    "Ptrack.h"

#define DEBUG	FALSE
#define VERBOSE_DEBUG	FALSE
#define GLOBAL


static void dprintf(char * /* s */, ...)
{
}

static inline double f_abs(double val)
{
    return (val < 0) ? -val : val;
}

static inline double long l_abs(long val)
{
    return (val < 0) ? -val : val;
}

static inline double long l_round(double val)
{
    return (val < 0) ? val - 0.5 : val + 0.5;
}

/*
 * Convert cents to ratio.
 */
static inline double cents_to_ratio(double cents)
{
    return pow(2.0, cents / 1200.0);
}

/*
 * Convert time constant (in milliseconds) to decay gain per sample.
 */
double tc_to_gain(double tc, long samprate)
{
    return exp(-1.0 / (0.001 * tc * samprate));
}

/*
 * Call to get the current period estimate. Return 0 if we don't
 * know what the pitch is. If we're tracking pitch, return the
 * current period estimate, otherwise if we're in a dropout, return
 * the last known period.
 */
double ptrack::get_period(void)
{
    return pitch_flag ? (track_flag ? period : last_period) : 0;
}

/*
 * Call to get current pitch.
 */
double ptrack::get_pitch(void)
{
double period;
    return ((period = get_period()) == 0) ? 0 : samprate / period;
}

/*
 * Initialize acquire mode.
 */
void ptrack::init_acquire()
{
    dprintf("ACQUIRE MODE\n");
    state = 0;
}

void ptrack::new_period_fn(void)
{
    if (output_fmt == FMT_ASCII) {
	fprintf(out_fp,"%ld %g\n", total,
		output_pitch ? get_pitch() : get_period());
    }
}

	
void ptrack::set_period(double period_param)
{
    period = period_param;
    new_period_fn();
}

/*
 * Initialize tracking mode with a period estimate.
 */
void ptrack::init_track(double new_period)
{
    dprintf("TRACK MODE, period %g\n",new_period);
    state = -1;
    set_period(new_period);
    n_events = 0;
    n_periods = 0;
    cur_delta = 0;
}

/*
 * Test for period dropouts and call note_on() if we have a new note.
 */
void ptrack::pitch_on(void)
{
    if (pitch_flag) {
	if (f_abs(period - last_period) > period * compare_ratio) {
	    note_off();
	    note_on();
	}
	else {
	    dprintf("dropout detected, pitch change %g cents\n",
		1200 * log(period / last_period) / log(2.0));
	}
    }
    else {
	pitch_flag = TRUE;
	note_on();
    }
}

/*
 * The dropout timer has count down to 0. Generate a note_off event.
 */
void ptrack::pitch_off(void)
{
    pitch_flag = FALSE;
    note_off();
    set_period(0);
}

/*
 * We've tracked enough periods for a valid pitch. Generate a
 * note_on event unless this is a pitch dropout.
 */
void ptrack::track_on(void)
{
    track_flag = TRUE;
    pitch_on();
}

/*
 * Latch the current period in last_period and set up the dropout
 * timer.
 */
void ptrack::track_off(void)
{
    if (track_flag) {
	track_flag = FALSE;
	last_period = period;
	dropout_timer = (long)(period * dropout_periods);
    }
}

/*
 * Process an event in tracking mode.
 */
void ptrack::proc_track(long delta)
{
    cur_delta += delta;
    /*
     * Is this event close to expected period?
     */
    if (f_abs(period - cur_delta) <= period * compare_ratio) {
#if VERBOSE_DEBUG
	dprintf("found event, cur_delta %ld period %g\n",
	    cur_delta,period);
#endif
	/*
	 * Update period estimate.
	 */
	set_period(period * (1 - track_gain) + cur_delta * track_gain);
	/*
	 * Update period count and compare to track_periods.
	 */
	if (++n_periods == track_periods)
	    track_on();
	cur_delta = 0;
	n_events = 0;
    }
    else {
	/*
	 * If we have too many events, or if we are past the period,
	 * go back to acquire mode. We can probably punt the max_events
	 * test and rely strictly on the cur_delta > period test.
	 */
	if (n_events++ > max_events) {
	    dprintf("too many events after %ld periods tracked\n",n_periods);
	    track_off();
	    init_acquire();
	}
	if (cur_delta > period) {
	    dprintf("missing event after %ld periods tracked\n",n_periods);
	    track_off();
	    init_acquire();
	}
	if (f_abs(period - 2 * cur_delta) <= period * compare_ratio) {
	    dprintf("possible octave error\n");
	    track_off();
	    init_acquire();
	}
    }
}

/*
 * Compare two periods, storing the average in the global compare_avg.
 * Return TRUE if periods are within compare_cents of each other, else FALSE.
 */
int ptrack::compare(long a, long b)
{
//    if (l_abs(a - b) <= l_round(a * compare_ratio)) 
// gcc on Linux has problems with above line
    if (labs(a - b) <= rint(a * compare_ratio)) {
	compare_avg = (double) (a + b) / 2;
	return TRUE;
    }
    else return FALSE;
}

/*
 * Call to compare two delta times. Go through the compare table
 * so we can generate reasonable debug messages and maintain
 * comparison statistics.
 */
int ptrack::compare_entry(int index, int state)
{
    int stat;
    char *desc;

    switch(index)  {
    case 0:
	desc = "d1,d2";
	stat = compare(d1, d2);
	break;
    case 1:
	desc = "d2,d3";
	stat = compare(d2, d3);
	break;
    case 2:
	desc = "d1+d2,d2+d3";
	if(ignore_d1pd2)	return FALSE;
	stat = compare(d1p2, d2p3);
	break;
    default:
	abort();
    }
    if (stat) {
	dprintf("successful compare(%s) in state %d\n", desc, state);
    }
    return stat;
	

/*
    if (ignore_d1pd2 && index == 2) return FALSE;
    stat = compare(*cmp_tab[index].p1, *cmp_tab[index].p2);
    if (stat) {
	dprintf("successful compare(%s) in state %d\n",
	    cmp_tab[index].name,state);
    }
 */

#ifdef JUSTACOMMENT
/*
 * The compare table is really used to generate nice debugging messages.
 * Each table entry defines a comparison between two
 * delta times, and associates a text string with the comparison.
 */
typedef struct {
    char *name;     /* name of comparison for debug messages */
    long *p1;	    /* pointer to first delta time value */
    long *p2;	    /* pointer to second delta time value */
} CMP_ENTRY;

CMP_ENTRY cmp_tab[] = {
{ "d1,d2", &d1, &d2 },
{ "d2,d3", &d2, &d3 },
{ "d1+d2,d2+d3", &d1p2, &d2p3 }
};

#define N_CMP	(sizeof(cmp_tab) / sizeof(CMP_ENTRY))

#endif /* JUSTACOMMENT */

}

/*
 * Process an event in acquire mode.
 */
void ptrack::proc_acquire(long delta)
{
    /*
     * If pitch_flag is TRUE, turn it off if acquire time exceeds
     * the dropout time.
     */
    if (pitch_flag) {
	dropout_timer -= delta;
	if (dropout_timer <= 0) pitch_off();
    }
    switch (state) {
    case 0:
	d1 = delta;
	state++;
	break;
    case 1:
	d2 = delta;
	d1p2 = d1 + d2;
	if (compare_entry(0,state)) {
	    init_track(compare_avg);
	}
	else state++;
	break;
    case 3:
	/*
	 * Shift all values and repeat state 2.
	 */
	d1 = d2;    /* actually, d1 isn't used in state 2 */
	d2 = d3;
	d1p2 = d2p3;
	state = 2;
	/*
	 * Falls through!
	 */
    case 2:
	d3 = delta;
	d2p3 = d2 + d3;
	if (compare_entry(1,state) ||
		compare_entry(2,state)) {
	    init_track(compare_avg);
	}
	else state++;
	break;
    }
}

/*
 * Process an event. Called from the xing event detector.
 */
void ptrack::accept_fn(long delta)
{
    /*
     * Remember this for rough period estimate.
     */
    last_delta = delta;
    /*
     * Process either track or acquire mode.
     */
    if (state < 0) proc_track(delta);
    else proc_acquire(delta);
}

/*
 * Open input and output files.
 */
void ptrack::open_files(char *inp_file, char *out_file)
{
char *read_fmt = "r";
char *write_fmt = "w";

    if (output_fmt == FMT_ASCII) write_fmt = "w";
    if (inp_file) {
	if (!strcmp(inp_file,"stdin")) inp_fp = stdin;
	else if ((inp_fp = fopen(inp_file,read_fmt)) == NULL) {
	    fprintf(stderr,"error, can't open file '%s'\n",inp_file);
	    exit(0);
	}
    }
    else {
	fprintf(stderr,"error, no input file specified\n");
	exit(0);
    }
    if (out_file) {
	if (!strcmp(out_file,"stdout")) out_fp = stdout;
	else if ((out_fp = fopen(out_file,write_fmt)) == NULL) {
	    fprintf(stderr,"error, can't create file '%s'\n",out_file);
	    exit(0);
	}
    }
    else {
	fprintf(stderr,"error, no output file specified\n");
	exit(0);
    }
}

/*
 * Close files.
 */
void ptrack::close_files(void)
{
    if (inp_fp && inp_fp != stdin) {
	fclose(inp_fp);
	inp_fp = NULL;
    }
    if (out_fp && out_fp != stdout) {
	fclose(out_fp);
	out_fp = NULL;
    }
}

/*
 * Call when done using ptrack.
 */
ptrack::~ptrack(void)
{
    close_files();
}

/*
 * Rounding.
 */
inline int s_round(double f)
{
    return (f > 0) ? (int)(f + 0.5) : (int)(f - 0.5);
}

/*
 * Call to process sample.
 */
void ptrack::proc_samp(short samp)
{
    xing::proc(samp);
}

/*
 * Process input file and write results to output file.
 */
void ptrack::proc_files(char *inp_file, char *out_file)
{
short samp;
double dval;
float fval;
short sval;
    open_files(inp_file, out_file);
    new_period_fn();
    while (fread(&samp,sizeof(short),1,inp_fp) == 1) {
	proc_samp(samp);
	total++;
	if (output_fmt != FMT_ASCII) {
	    dval = output_pitch ? get_pitch() : get_period();
	    if (output_fmt == FMT_SHORT) {
		sval = s_round(dval * output_scale);
		if (fwrite(&sval,sizeof(short),1,out_fp) != 1) {
		    fprintf(stderr,"error writing to output file\n");
		    exit(0);
		}
	    }
	    else {
		fval = dval;
		if (fwrite(&fval,sizeof(float),1,out_fp) != 1) {
		    fprintf(stderr,"error writing to output file\n");
		    exit(0);
		}
	    }
	}
    }
    new_period_fn();
    close_files();
}

/*
 * Set up default pitch track parameters.
 */
void ptrack::default_params(void)
{
//    samprate   = 44100;
    output_scale = 1.0;
    max_events   = 1;
    track_gain   = 0.5;
//    compare_cents   = 200;
    SetCompareCents(200);
    dropout_periods = 4;
    /*
     * track_periods is number of tracked periods that need to
     * occur before we accept the pitch as valid. Since we
     * have already looked at two periods in acquire mode, the
     * total number of periods required to determine a note will
     * be track_periods + 2.
     */
//    track_per_ui = 3;
    SetTrackPerUI(3);
    ignore_d1pd2 = TRUE;
    xing::default_params();
}

/*
 * Set up various pitch tracking algorithms.
 */
void ptrack::music_params()
{
    default_params();
//    compare_cents = 100;
    SetCompareCents(100);
}

void ptrack::speech_params()
{
    default_params();
//    compare_cents = 200;
    SetCompareCents(200);
}

/*
 * Query user for pitch track parameters.
 */
void ptrack::query_params()
{
    xing::query_params();
#ifdef BILLG
    getint("number of in-between events allowed per period",&max_events);
    getdouble("period tracking gain",&track_gain);
    getdouble("tracking threshold (cents)",&compare_cents);
    getint("number dropout periods allowed",&dropout_periods);
    getint("number of periods for valid note (2, 3, or 4)",&track_per_ui);
#endif /* BILLG */
}

void ptrack::SetCompareCents(double compCs)
{
    compare_cents = compCs;
    compare_ratio = cents_to_ratio(compare_cents) - 1.0;
}

void ptrack::SetTrackPerUI(int tks)
{
    track_per_ui = tks;
    track_periods = track_per_ui - 2;
}

/*
 * Initialize the pitch tracker.
 */
ptrack::ptrack(long srate) : xing(srate)
{
    /*
     * Reset the important stuff.
     */
    total = 0;
    period = 0;
    track_flag = FALSE;
    pitch_flag = FALSE;
    /*
     * Start in ACQUIRE mode.
     */
    init_acquire();
    /*
     * Initialize zero crossing detector.
     */
    // initialize global fields to something reasonable - AFTER Xing set
    default_params();
}

