// -*- C++ -*-

// tclTemplate.h
//
// Template to define a tcl interface to an object
// in this case, specialized for objects that are the 
// particular elements of the Audio Pattern Recognition Language
// environment.
// 
// dpwe@media.mit.edu 1995feb12
//

// To use this, you set up several definitions, then just include 
// the `header' file to provide the rest of the code.  E.g., you must
// #define the following symbols -
//
//  OBJECT        - The C++ object involved, e.g.   SoundSrc
//  TCLNAME       - the name of the new tcl command, e.g. soundsrc
//  STRUCTFIELDS  - inserted into the parent struct definition
//  CONFIGLINES   - inserted into the config table 
//                  (good to use CONFIGLINE(TKTYPE, FIELD, CAPFIELD, DFLT))
//  COMMANDCODE   - inserted into widget command function
//  COMMANDLIST   - lists extra commands for error message
//  CONSTRUCTCODE - how to call the constructor
//  CONFIGCODE    - extra configuration calls to install extra args
//  TYPE          - as one of SINK, SOURCE or PIPE, which controls 
//                    whether the object has an INPUT field, and 
//                    whether it is registered in the datasrc hash table.
//
// See the example tclCompPipe.cc for more details.  In particular, 
// there is a nasty leading 'else' required for COMMANDCODE, and 
// a leading "/" separator preferred for COMMANDLIST
//
// To link the new commands into the tcl interpreter, you will 
// need to call <TCLNAME>_Init(interp) from the block of such calls 
// in tkAppInit.c .  After that, you should be all set.
//
// -------------------------------------------------------------------------

/* construct all the symbols we need */

#ifndef TCLNAMER	// first time we have been included
#  define ___quote(ARG)      #ARG
#  define ___QUOTE(ARG)	___quote(ARG)
#  define CONFIGLINE(TKTYPE, FIELD, CAPFIELD, DEFAULT)	\
	  {TKTYPE, "-" ___QUOTE(FIELD), ___QUOTE(FIELD), ___QUOTE(CAPFIELD), \
	   DEFAULT, Tk_Offset(TCLSTRUCT, FIELD), 0},

  // You can build up lines of a config table like
  // 
  //   CONFIGLINE(TK_CONFIG_DOUBLE, reference, Reference, "1.0")
  //
  // as long as 'reference' is both the config name and the 
  // name of the field within the tcl struct, and TCLSTRUCT 
  // will be defined as the name of the struct (it is, in tclTemplate.h)
  // 

#  define ___cat(ARG1,ARG2)	ARG1 ## ARG2
#  define ___CAT(ARG1,ARG2)	___cat(ARG1,ARG2)

#  include "tclAprlMisc.h"

#  define FIRSTTIME	// flag for other things that might vary 

#else /* TCLNAMER is defined - must be subsequent time we are called */
#  undef TCLNAMER
#  undef TCLNAMEQ
#  undef TCLSTRUCT
#  ifdef FIRSTTIME
#    undef FIRSTTIME
#  endif /* FIRSTTIME */
#endif /* TCLNAMER */

#define TCLNAMER(REST)	___CAT(TCLNAME,REST)

// Now, when you define TCLNAME in your header file, the tclTemplate's 
// use of TCLNAMER(rest) will get the remainder correctly.

#define TCLNAMEQ       ___QUOTE(TCLNAME)

#define TCLSTRUCT TCLNAMER(Struct)

#ifdef ISPIPE
#define ISSINK
#define ISSOURCE
#endif /* ISPIPE */

#ifdef ISCOMB
#define ISSINK
#define ISSOURCE
#endif /* ISCOMB */

// tcl object data structure

typedef struct {
    Tcl_Interp 	*interp;
    Tk_Window	tkwin;		// fake, needed for Tk_Config
#ifdef ISSINK
    char	*input;
#endif /* ISSINK */
#ifdef ISCOMB
    char 	*input2;
#endif /* ISCOMB */
    STRUCTFIELDS
    OBJECT	*obj;		// the actual C++ object
} TCLSTRUCT;

// Config options

static Tk_ConfigSpec TCLNAMER(ConfigSpecs)[] = {
    CONFIGLINES
#ifdef ISSINK
    {TK_CONFIG_STRING, "-input", "input", "Input",
	NONESTR, Tk_Offset(TCLSTRUCT, input), 0},
#endif /* ISSINK */
#ifdef ISCOMB
    {TK_CONFIG_STRING, "-in2", "in2", "In2",
	NONESTR, Tk_Offset(TCLSTRUCT, input2), 0},
#endif /* ISCOMB */
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

extern "C" {

extern char *strdup(const char *s1);	// difficult to get this declared

// tcl interface function definitions

int TCLNAMER(_Init)(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
}

static int TCLNAMER(Command)(ClientData clientData, Tcl_Interp *interp, 
		   int argc, char **argv);
static int TCLNAMER(Proc)(ClientData clientData, Tcl_Interp *interp, 
		int argc, char **argv);
static int TCLNAMER(Configure)(Tcl_Interp *interp, TCLSTRUCT *as, 
			    char *cmdName, int argc, char **argv, int flag);
static void TCLNAMER(Delete)(ClientData clientData);

int TCLNAMER(_Init)(Tcl_Interp *interp)
{   /* add the tcl command defined in this package to the interpreter */
    Tcl_CreateCommand(interp, TCLNAMEQ, TCLNAMER(Command), (ClientData) NULL,
		      (void (*)(ClientData)) NULL);
    return TCL_OK;
}

static int TCLNAMER(Command)(ClientData /* clientData */, Tcl_Interp *interp, 
			     int argc, char **argv)
{	/* tcl command to instance a new TCLNAMER object */
    char cmdName[32];
    if(argc < 2) {
	Tcl_AppendResult(interp, "wrong # args; should be \"", 
			 argv[0], " objName ?options?\"", NULL);
	return TCL_ERROR;
    }
    // process for a command name
    if(strcmp(argv[1], ".")==0)  {
	// "." is a magic name, meaning just make up a unique one for me
	strcpy(cmdName, UniqueSPName(interp));
    } else if(argv[1][0] == '-')  {
	// if the name starts with '-', it's probably a config option 
	// by mistake - we won't have it.
	Tcl_AppendResult(interp, argv[0],": \"", argv[1], 
			 "\" is not an acceptable object name", NULL);
	return TCL_ERROR;
    } else {
	strcpy(cmdName, argv[1]);
    }

    // allocate a new object
    TCLSTRUCT *ptr = new TCLSTRUCT;
    memset(ptr,0,sizeof(TCLSTRUCT));
    ptr->interp = interp;
    ptr->tkwin  = Tk_MainWindow(interp);   // dummy for Tk_Config
    ptr->obj   = 0;	/* don't create object until configure */
#ifdef ISSINK
    ptr->input  = 0;	// critical, else frees random memory
#endif /* ISSINK */
    // put it into the tcl command namespace
    Tcl_CreateCommand(interp, cmdName, TCLNAMER(Proc), 
		      (ClientData)ptr, TCLNAMER(Delete));
    // pass rest of command to configure
    if(TCLNAMER(Configure)(interp,ptr,cmdName,argc-2,argv+2,0)
       !=TCL_OK) {
	return TCL_ERROR;
    }
    Tcl_AppendResult(interp, cmdName, NULL);
    return TCL_OK;
}

static int TCLNAMER(Proc)(ClientData clientData, Tcl_Interp *interp, 
		int argc, char **argv)
{	/* handle the commands supported by the tcl object */
    TCLSTRUCT *ptr = (TCLSTRUCT *)clientData;
    int		result = TCL_OK;
    char	c;
    int		length;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " <command> ?args?\"", NULL);
	return TCL_ERROR;
    }
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, ptr->tkwin, TCLNAMER(ConfigSpecs), 
				      (char *)ptr, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, ptr->tkwin, TCLNAMER(ConfigSpecs), 
				      (char *)ptr, argv[2], 0);
	} else {
	    result =TCLNAMER(Configure)(interp, ptr, argv[0], argc-2, argv+2, 
				       TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, ptr->tkwin, TCLNAMER(ConfigSpecs),
				   (char *) ptr, argv[2], 0);
    } 
#ifdef ISSOURCE
#define SRCCOMMS  "/more/sampleRate/time/value"
#define SNKCOMMS ""
    else if ( (c=='m') && (strncmp(argv[1], "more", length)==0) )  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], "more\"", NULL);
	    result = TCL_ERROR;
	} else {
	    char s[32];
	    sprintf(s, "%d", ptr->obj->ThereIsMore());
	    Tcl_AppendResult(interp, s, NULL);
	}
    }
    else if ( (c=='s') && (strncmp(argv[1], "sampleRate", length)==0) )  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], "sampleRate\"", NULL);
	    result = TCL_ERROR;
	} else {
	    char s[32];
	    sprintf(s, "%.1f", ptr->obj->SampleRate());
	    Tcl_AppendResult(interp, s, NULL);
	}
    }
    else if ( (c=='t') && (strncmp(argv[1], "time", length)==0) )  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], "time\"", NULL);
	    result = TCL_ERROR;
	} else {
	    char s[32];
	    sprintf(s, "%.3f", ptr->obj->CountTime());
	    Tcl_AppendResult(interp, s, NULL);
	}
    }
    else if ( (c=='v') && (strncmp(argv[1], "value", length)==0) )  {
	if(argc!=2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], "value\"", NULL);
	    result = TCL_ERROR;
	} else {
	    char s[32];
	    float f;
	    ptr->obj->readFrom(&f, 1, ptr->obj->Count()-1);
	    sprintf(s, "%.3f", f);
	    Tcl_AppendResult(interp, s, NULL);
	}
    }
#else /* !ISSOURCE */
#define SRCCOMMS ""
#endif /* ISSOURCE */
    COMMANDCODE(interp,ptr,c,argc,argv,length)
    else {
	Tcl_AppendResult(interp, "bad option \"", argv[1], 
			 "\":  configure/cget" SRCCOMMS COMMANDLIST,
			 (char *) NULL);
#undef SRCCOMMS
	result = TCL_ERROR;
    }
    return result;
}

static int TCLNAMER(Configure)(Tcl_Interp *interp, TCLSTRUCT *ptr, 
			    char *cmdName, int argc, char **argv, int flags)
{	/* Handle arguments specifically to configure */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, ptr->tkwin, TCLNAMER(ConfigSpecs), 
			  argc, argv, (char *)ptr, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
#ifdef ISSINK
    DataSource *in = 0;

    if( (rslt = DSfromString(interp, &(ptr->input), &in, 
			     !ConfigSpecified("-input", argc, argv)))
         != TCL_OK)
	return rslt;
    
#endif /* ISSINK */
#ifdef ISCOMB
    DataSource *in2;
    if( (rslt = DSfromString(interp, &(ptr->input2), &in2, 0))
         != TCL_OK)
	return rslt;
#endif /* ISCOMB */
    if(ptr->obj == 0)  {
#ifdef ISCOMB
	CONSTRUCTCODE(ptr, in, in2);
#else /* !ISCOMB */
#ifdef ISSINK
	CONSTRUCTCODE(ptr, in);
#else
	CONSTRUCTCODE(ptr);
#endif /* ISSINK */
#endif /* !ISCOMB */
	// schedule it
#ifdef ISSINK
	if(in != 0) {
#endif /* ISSINK */
	    if(globalScheduler==0)  {
		fprintf(stderr,TCLNAMEQ"Config:Error: no scheduler defined\n");
	    } else {
		globalScheduler->Enschedule(ptr->obj);
	    }
#ifdef ISSINK
	}
#endif /* ISSINK */
#ifdef ISSOURCE
	// Register the DataSource in the hash table
	int isnew;
	Tcl_HashEntry *entryPtr;
	entryPtr = Tcl_CreateHashEntry(&datasrcTable, cmdName, &isnew);
	Tcl_SetHashValue(entryPtr, (DataSource *)(ptr->obj));
	// use this source as the next default input
	globalDataSource = (DataSource *)(ptr->obj);
	if(globalDSnameStr)	ckfree(globalDSnameStr);
	globalDSnameStr  = strdup(cmdName);
#endif /* ISSOURCE */
    } else {
#ifdef ISSINK
	DataSource *oldin = ptr->obj->Input();
	if(in != oldin)  {
	    ptr->obj->SetInput(in);
	    if(globalScheduler==0)  {
		fprintf(stderr,TCLNAMEQ "Config:Error:no scheduler defined\n");
	    } else {
		if(in==0)  
		    globalScheduler->Deschedule(ptr->obj);
		else if(oldin==0)
		    globalScheduler->Enschedule(ptr->obj);
	    }
	}
#endif /* ISSINK */
#ifdef ISCOMB
	DataSource *oldin2 = ptr->obj->Input2();
	if(in2 != oldin2)  {
	    ptr->obj->SetInput2(in2);
	}
#endif /* ISCOMB */
	CONFIGCODE(ptr);
    } 
    return rslt;
}

static void TCLNAMER(Delete)(ClientData clientData)
{	/* a particular obj is being deleted */
    TCLSTRUCT *ptr = (TCLSTRUCT *)clientData;
    if(ptr->obj)
	delete ptr->obj;
    delete ptr;
#ifdef ISSOURCE
    /* we really want to unregister the hash table entry here too. */
    int deleted = 0;
    Tcl_HashSearch hs;
    Tcl_HashEntry *hePtr;
    hePtr = Tcl_FirstHashEntry(&datasrcTable, &hs);
    while(hePtr)  {
	if(Tcl_GetHashValue(hePtr)==(ClientData)(ptr->obj))  {  // found it
	    Tcl_DeleteHashEntry(hePtr);
	    deleted = 1;
	    hePtr=0;
	} else {
	    hePtr = Tcl_NextHashEntry(&hs);
	}
    }
    if(!deleted)  {
	fprintf(stderr,TCLNAMEQ "Delete:unable to dehash 0x%lx\n", clientData);
    }
#endif /* ISSOURCE */
}
    
// Before leaving, undefine all the symbols that might be redefined 
// if a source file were to try and include us again

#undef OBJECT
#undef TCLNAME
#ifdef ISSOURCE
#undef ISSOURCE
#endif /* ISSOURCE */
#ifdef ISSINK
#undef ISSINK
#endif /* ISSINK */
#ifdef ISPIPE
#undef ISPIPE
#endif /* ISPIPE */
#ifdef ISCOMB
#undef ISCOMB
#endif /* ISCOMB */
#undef STRUCTFIELDS
#undef CONFIGLINES
#undef COMMANDCODE
#undef COMMANDLIST
#undef CONSTRUCTCODE
#undef CONFIGCODE
