/*
 * bread.C
 * 
 * Tcl extensions for reading binary streams
 *
 * 1996oct01 dpwe@icsi.berkeley.edu
 * $Header: /u/drspeech/src/SPRACHcore/guitools/dpweutils_tcl/RCS/bread.C,v 1.5 1998/12/03 00:04:37 dpwe Exp $
 */

/*********************** NOTE ******************************************
 * Under Tcl7.5, the default translation mode for stdin is 'auto', 
 * which works fine for most binary streams UNLESS the sequence 
 * 0x0d 0x0a occurs, in which case it is translated to 0x0a (losing 
 * one byte);  Of course, one-in-65536 of randomly-chosen floats 
 * will contain this sequence, throwing the stream out of step for 
 * the rest of the input...
 *
 * IF YOU'RE READING BINARY FLOATS FROM STDIN REMEMBER THE FOLLOWING:
 * 
 * 	fconfigure stdin -translation binary
 *
 * You probably don't want line-based buffering either, so try
 *
 * 	fconfigure stdin -buffering none
 * or
 * 	fconfigure stdin -buffering full
 * 
 * see "man fconfigure"
 *
 * 1996oct04 dpwe@icsi.berkeley.edu
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>	/* for malloc, inside ckalloc */
#include <string.h>
#include <assert.h>
#include <tcl.h>	/* also defines EXTERN in C++-friendly manner */

/* Forward function definitions */
EXTERN int Tcl_breadiCmd(ClientData clientData, Tcl_Interp *interp, 
			 int argc, char **argv);
EXTERN int Tcl_breadfCmd(ClientData clientData, Tcl_Interp *interp, 
			 int argc, char **argv);
EXTERN int Tcl_bwriteiCmd(ClientData clientData, Tcl_Interp *interp, 
			  int argc, char **argv);
EXTERN int Tcl_bwritefCmd(ClientData clientData, Tcl_Interp *interp, 
			  int argc, char **argv);

EXTERN int Tcl_bfilecopyCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv);

/* inline */
inline int min(int a, int b) {
    int v = b;
    if (a<b)	v = a;
    return v;
}

inline int max(int a, int b) {
    int v = b;
    if (a>b)	v = a;
    return v;
}

/* required function name different on SGI */
#ifdef _NO_INITNAME_TAGS
/* SGI style */
#define Dpweutils_tcl_Init(a) 		Dpweutils_Init(a)
#define Dpweutils_tcl_SafeInit(a)	Dpweutils_SafeInit(a)
#endif /* _NO_INITNAME_TAGS */

/* Initialization commands */
EXTERN int Dpweutils_tcl_Init(Tcl_Interp *interp)
{   /* Package opening call */
    
    /* for Tcl7.5+, signal the package mechanism */
    Tcl_PkgProvide(interp, "Dpweutils_tcl", DTUTILS_VERSION);

    Tcl_CreateCommand(interp, "breadi", Tcl_breadiCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    Tcl_CreateCommand(interp, "breadf", Tcl_breadfCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    Tcl_CreateCommand(interp, "bwritei", Tcl_bwriteiCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    Tcl_CreateCommand(interp, "bwritef", Tcl_bwritefCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    Tcl_CreateCommand(interp, "bfilecopy", Tcl_bfilecopyCmd, 
		      (ClientData) NULL, (void (*)(ClientData)) NULL);
    return TCL_OK;
}

EXTERN int Dpweutils_tcl_SafeInit(Tcl_Interp *interp)
{   /* Package opening call */
/*    fprintf(stderr, "Dpweutils_tcl: cannot load under safe interpreter\n");
    return TCL_ERROR; */
    return Dpweutils_tcl_Init(interp);
}

/* Reading non-ascii input streams within Tcl is a real bind because
 * the streams are basically treated as ascii and you have to do the 
 * conversion to numbers algorithmically.  But the worst part is if 
 * the number contains a zero byte, since this marks the end of the 
 * read string, so you have to do it byte-by-byte, then check for 
 * zero-length strings and interpret that as a byte of zero.  Definite 
 * case for a C extension.
 */


/*
 *----------------------------------------------------------------------
 *
 * Tcl_breadiCmd --
 *
 *	Tcl command to read a binary-encoded number from an input stream.
 *	Command syntax is: 
 *		breadi stream {nbytes 4} {bigendian 1}
 *
 * Results:
 *      Sets interp->result to an ascii string representing the number 
 * 	that was read from the stream.  If EOF is hit, the routine 
 *      completes OK, but an empty string is returned.
 *
 * Side effects:
 *	Input file pointer is advanced by <nbytes> bytes
 *
 *----------------------------------------------------------------------
 */

int Tcl_breadiCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to read from. */
    int mode;				/* Mode in which channel is opened. */
    int totalread, thistime;		/* byte counters for transfer */
    unsigned char buf[8];		/* space to read into */
    long num = 0;			/* the answer */
    int  nbytes = 4;			/* the number of bytes to read */
    int  bigendian = 1;			/* flag to treat data as bigendian */
    int  i;
    long v;

    if(argc < 2 || argc > 4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream {nbytes 4} {bigendian 1}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 2) {
	if (Tcl_GetInt(interp, argv[2], &nbytes) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if(argc > 3) {
	if (Tcl_GetInt(interp, argv[3], &bigendian) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Transfer the bytes */
    thistime = -1;
    totalread = 0;
    while(totalread < nbytes && thistime != 0) {
	thistime = Tcl_Read(chan, (char *)&(buf[totalread]), nbytes-totalread);
	if (thistime < 0) {
            Tcl_AppendResult(interp, "error reading \"",
			     Tcl_GetChannelName(chan), "\": ",
			     Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
	}
	totalread += thistime;
    }
    
    if(totalread < nbytes) {	/* Assume we hit eof */
	Tcl_AppendResult(interp, "", NULL);
	return TCL_OK;
    }

    /* OK, we got the desired number of bytes into buf */
    for(i = 0; i < nbytes; ++i) {
	if (bigendian) {
	    v = buf[i];
	} else {	/* littleendian */
	    v = buf[nbytes - 1 - i];
	}
	if(i == 0 && v > 127) {  /* Make leading byte signed */
	    v = v - 256;
	}
	num = 256*num + v;
    }

    /* num  is the integer read - convert to ascii & return */
    sprintf(interp->result, "%ld", num);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_bwriteiCmd --
 *
 *	Tcl command to  a binary-encoded number from an input stream.
 *	Command syntax is: 
 *		bwritei stream val {nbytes 4} {bigendian 1}
 *
 * Results:
 *      Sets interp->result to the number of bytes written to the stream.
 *
 * Side effects:
 *	Output file pointer is advanced by <nbytes> bytes, data written out.
 *
 *----------------------------------------------------------------------
 */

int Tcl_bwriteiCmd(ClientData clientData, Tcl_Interp *interp, 
		   int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to write to. */
    int mode;				/* Mode in which channel is opened. */
    unsigned char buf[8];		/* space to read into */
    long num = 0;			/* the answer */
    int  nbytes = 4;			/* the number of bytes to read */
    int  bigendian = 1;			/* flag to treat data as bigendian */
    int  val;

    if(argc < 3 || argc > 5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream val {nbytes 4} {bigendian 1}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 3) {
	if (Tcl_GetInt(interp, argv[3], &nbytes) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if(argc > 4) {
	if (Tcl_GetInt(interp, argv[4], &bigendian) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    /* Get the actual value */
    if (Tcl_GetInt(interp, argv[2], &val) != TCL_OK) {
	return TCL_ERROR;
    }

    /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Transfer the bytes */
    {
	int writ = 0, thistime = -1, towrite;
	int ii = 0, iz = sizeof(int);
	int i,j;
	int intbytes = min(sizeof(int), nbytes);

	if(bigendian) {	/* pad with zeros first */
	    while(thistime != 0 && writ + intbytes < nbytes) {
		towrite = min(iz, nbytes - intbytes - writ);
		thistime = Tcl_Write(chan, (char*)&ii, towrite);
		writ += thistime;
	    }
	}
	/* write out the bottom nbytes of val */
	for(i = 0; i<intbytes; ++i) {
	    if(bigendian) {	/* write MSB first */
#ifdef WORDS_BIGENDIAN
		j = sizeof(int) - intbytes + i;
#else /* !WORDS_BIGENDIAN */
		j = intbytes - 1 - i;
#endif /* WORDS_BIGENDIAN */
	    } else {	/* little-endian -> write LSB first */
#ifdef WORDS_BIGENDIAN
		j = sizeof(int) - 1 - i;
#else /* !WORDS_BIGENDIAN */
		j = i;
#endif /* WORDS_BIGENDIAN */
	    }
	    thistime = Tcl_Write(chan, ((char*)&val)+j, 1);
	    writ += thistime;
	    if(thistime == 0)  break;
	}
	if(!bigendian) {	/* pad with zeros *afterwards* */
	    while(thistime != 0 && writ < nbytes) {
		towrite = min(iz, nbytes - writ);
		thistime = Tcl_Write(chan, (char*)&ii, towrite);
		writ += thistime;
	    }
	}
	num = writ;
    }

    /* return number of bytes written */
    sprintf(interp->result, "%ld", num);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_breadfCmd --
 *
 *	Tcl command to read a binary-encoded float from an input stream.
 *	Command syntax is: 
 *		breadf stream {nbytes 4} {bigendian 1}
 *
 * Results:
 *      Sets interp->result to an ascii string representing the decimal 
 * 	number that was read from the stream.  If EOF is hit, the routine 
 *      completes OK, but an empty string is returned.
 *	nbytes is 4 for floats or 8 for doubles; other values give an 
 *      error.
 *
 *	This assumes sizeof(float)==4 and sizeof(double)==8.  
 *	sizeof is used instead of constants in the code.
 *
 * Side effects:
 *	Input file pointer is advanced by $nbytes bytes
 *
 *----------------------------------------------------------------------
 */

int Tcl_breadfCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to read from. */
    int mode;				/* Mode in which channel is opened. */
    int totalread, thistime;		/* byte counters for transfer */
    unsigned char *buf;			/* space to read into */
    int  nbytes = sizeof(float);	/* the number of bytes to read */
    int  bigendian = 1;			/* flag to treat data as bigendian */
#ifdef WORDS_BIGENDIAN
    int  hostbigendian = 1;
#else        /* Must be LITTLEENDIAN */
    int  hostbigendian = 0;
#endif /* BIG_ENDIAN */
    int  i;
    long v;
    float fnum = 0;			/* the answer */
    double dnum = 0;

    if(argc < 2 || argc > 4) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream {nbytes 4} {bigendian 1}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 2) {
	if (Tcl_GetInt(interp, argv[2], &nbytes) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if(argc > 3) {
	if (Tcl_GetInt(interp, argv[3], &bigendian) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    /* point the pointer to one of the num vars dep on nbytes */
    if (nbytes == sizeof(fnum)) {
	buf = (unsigned char *)&fnum;
    } else if (nbytes == sizeof(dnum)) {
	buf = (unsigned char *)&dnum;
    } else {
	Tcl_AppendResult(interp, argv[0], ": nbytes of ", argv[2], 
			 " is not sizeof(float) or sizeof(double)", NULL);
	return TCL_ERROR;
    }
	
   /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Transfer the bytes */
    thistime = -1;
    totalread = 0;
    while(totalread < nbytes && thistime != 0) {
	thistime = Tcl_Read(chan, (char *)&(buf[totalread]), nbytes-totalread);
	if (thistime < 0) {
            Tcl_AppendResult(interp, "error reading \"",
			     Tcl_GetChannelName(chan), "\": ",
			     Tcl_PosixError(interp), (char *) NULL);
            return TCL_ERROR;
	}
	totalread += thistime;
    }
    
    if(totalread < nbytes) {	/* Assume we hit eof */
	Tcl_AppendResult(interp, "", NULL);
	return TCL_OK;
    }

    /* OK, we got the desired number of bytes into buf */
    /* reverse the bytes if mode is not native */
    if(bigendian != hostbigendian) {
	for(i=0; i<nbytes/2; ++i) {
	    v = buf[i]; buf[i] = buf[nbytes - 1 - i]; buf[nbytes - 1 - i] = v;
	}
    }

    /* num  is the integer read - convert to ascii & return */
    if (nbytes == sizeof(float)) {
	sprintf(interp->result, "%.6f", fnum);
    } else {	/* assumed double */
	sprintf(interp->result, "%.6f", dnum);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_writefCmd --
 *
 *	Tcl command to write a binary-encoded float to an output stream.
 *	Command syntax is: 
 *		bwritef stream val {nbytes 4} {bigendian 1}
 *
 * Results:
 *      Sets interp->result to the number of bytes written (0 on error).
 *
 * Side effects:
 *      File pointer is advanced by 4 bytes; data is written.
 *
 *----------------------------------------------------------------------
 */

int Tcl_bwritefCmd(ClientData clientData, Tcl_Interp *interp, 
		  int argc, char **argv)
{
    Tcl_Channel chan;			/* The channel to read from. */
    int mode;				/* Mode in which channel is opened. */
    int  nbytes = sizeof(float);	/* the number of bytes to read */
    int  bigendian = 1;			/* flag to treat data as bigendian */
#ifdef WORDS_BIGENDIAN
    int  hostbigendian = 1;
#else        /* Must be LITTLE_ENDIAN */
    int  hostbigendian = 0;
#endif /* BIG_ENDIAN */
    float fnum = 0;			/* the number */
    double dnum = 0;			/* alternate number */
    unsigned char *buf;			/* space to write from */
    int  writ = 0;

    if(argc < 3 || argc > 5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " stream val {nbytes 4} {bigendian 1}\"", NULL);
	return TCL_ERROR;
    }

    /* get the float */
/*    if (Tcl_GetFloat(interp, argv[2], &val) != TCL_OK) {
	return TCL_ERROR;
    } */
    /* no type checking on input float, sorry */
    fnum = dnum = atof(argv[2]);

    /* handle optional arguments */
    if(argc > 3) {
	if (Tcl_GetInt(interp, argv[3], &nbytes) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if(argc > 4) {
	if (Tcl_GetInt(interp, argv[4], &bigendian) != TCL_OK) {
	    return TCL_ERROR;
	}
    }

    /* point the pointer to one of the num vars dep on nbytes */
    if (nbytes == sizeof(fnum)) {
	buf = (unsigned char *)&fnum;
    } else if (nbytes == sizeof(dnum)) {
	buf = (unsigned char *)&dnum;
    } else {
	Tcl_AppendResult(interp, argv[0], ": nbytes of ", argv[3], 
			 " is not sizeof(float) or sizeof(double)", NULL);
	return TCL_ERROR;
    }
	
    /* Figure out the channel implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    chan = Tcl_GetChannel(interp, argv[1], &mode);
    if (chan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Transfer the bytes */
    {
	int thistime = -1, towrite;
	int i,j;
	for(i = 0; i<nbytes; ++i) {
	    if(bigendian == hostbigendian) {
		j = i;
	    } else {
		j = nbytes - 1 - i;
	    }
	    thistime = Tcl_Write(chan, (char*)(buf+j), 1);
	    writ += thistime;
	    if(thistime == 0)  break;
	}
    }

    /* return number of bytes written */
    sprintf(interp->result, "%d", writ);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_bfilecopyCmd --
 *
 *	Tcl command to copy binary data between file streams
 *	Command syntax is: 
 *		bfilecopy instream outstream {nbytes -1} {stopstring ""}
 *
 * Results:
 *      returns the actual number of bytes transferred, or -1 if it 
 *      couldn't transfer.
 *
 * Side effects:
 *	Binary data is transferred from the instream to the outstream, 
 *      advancing both file pointers.  Up to $nbytes of data are 
 *	transferred (which is not a limit if $nbytes has its default 
 *	negative value).  Transfer will stop at EOF of input stream, 
 *	OR if the sequence $stopstring is encountered in the stream 
 *	(unless it is an empty string).  If transfer is terminated by 
 * 	encountering $stopstring, that string is NOT copied to the output, 
 * 	but it IS consumed from the input stream.
 *
 *----------------------------------------------------------------------
 */

static int _bfc_readsome(char *buf, Tcl_Channel& inchan, int& thistime, 
			 int& remain, int bufsize, int& totalread, 
			 Tcl_Interp* interp)
{   /* subsidiary function to read some data into a buffer & update counters */
    thistime = Tcl_Read(inchan, buf, 
			(remain<0)?bufsize:min(remain, bufsize));
    if (thistime < 0) {
	Tcl_AppendResult(interp, "error reading \"",
			 Tcl_GetChannelName(inchan), "\": ",
			 Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }
    if(remain > 0)
	remain -= thistime;
    totalread += thistime;
/*    fprintf(stderr, "@%d: read '%s'\n", totalread, buf); */
    return TCL_OK;
}

static int _bfc_writesome(char *buf, Tcl_Channel& outchan, int thistime, 
			  int& totalwrit, Tcl_Interp* interp)
{   /* subsidiary function to write out some data */
    int did;
    did = Tcl_Write(outchan, buf, thistime);
    if (did < thistime) {
	Tcl_AppendResult(interp, "error writing \"",
			 Tcl_GetChannelName(outchan), "\": ",
			 Tcl_PosixError(interp), (char *) NULL);
	return TCL_ERROR;
    }
    totalwrit += did;
    return TCL_OK;
}

#define BCPY_BUFSIZE 1024

int Tcl_bfilecopyCmd(ClientData clientData, Tcl_Interp *interp, 
		     int argc, char **argv)
{
    Tcl_Channel inchan;			/* The channel to read from. */
    Tcl_Channel outchan;		/* The channel to write to. */
    int mode;				/* Mode in which channel is opened. */
    int totalread, thistime;		/* byte counters for transfer */
    char *buf;				/* space to read into */
    int  nbytes = -1;			/* the number of bytes to read */
    int  i;
    long v;

    int bufsize;
    int remain;
    int sslen = 0;
    int seenstop = 0;
    int totalwrit = 0;
    int thistimepart;

    char *stopstring = NULL;
    int toMatch, matchAt;
    int len;

    if(argc < 3 || argc > 5) {
	Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], 
			 " instream outstream {nbytes -1} {stopstring \"\"}\"", NULL);
	return TCL_ERROR;
    }

    /* handle optional arguments */
    if(argc > 3) {
	if (Tcl_GetInt(interp, argv[3], &nbytes) != TCL_OK) {
	    return TCL_ERROR;
	}
    }
    if(argc > 4) {
	if (strlen(argv[4]) > 0) {
	    stopstring = argv[4];
	}
    }

    /* Figure out the channels implicated */
    /* This code is copied from tcl7.5/generic/tclIOCmd.c:Tcl_GetsCmd .. */
    inchan = Tcl_GetChannel(interp, argv[1], &mode);
    if (inchan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_READABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[1],
                "\" wasn't opened for reading", (char *) NULL);
        return TCL_ERROR;
    }
    
    outchan = Tcl_GetChannel(interp, argv[2], &mode);
    if (outchan == (Tcl_Channel) NULL) {
	return TCL_ERROR;
    }
    if ((mode & TCL_WRITABLE) == 0) {
        Tcl_AppendResult(interp, "channel \"", argv[2],
                "\" wasn't opened for writing", (char *) NULL);
        return TCL_ERROR;
    }
    
    /* Transfer the bytes */
    /* because we want to read just to the end of stopstring, but no 
       further, we cannot read more than strlen(stopstring) bytes at a 
       time - unless there is no stopstring. */
    bufsize = (stopstring==NULL)?BCPY_BUFSIZE:strlen(stopstring);
    buf = (char*)ckalloc(bufsize);
    if (stopstring) {sslen = strlen(stopstring);}    

    thistime = -1;
    remain = nbytes;
    while(remain != 0 && thistime != 0 && seenstop == 0) {
	if(_bfc_readsome(buf, inchan, thistime, remain, bufsize, totalread, interp)
	   == TCL_ERROR) {
	    return TCL_ERROR;
	}
	if(sslen) {  /* search for stopstring */
	    int matchAt = -1;
	    len = sslen;
	    /* search over all start positions for stopstring */
	    i = 0;
	    while(i < thistime) {
		int initialmatchable = min(thistime - i, len);
		if(strncmp(buf+i, stopstring, initialmatchable) == 0) {
		    matchAt = i;
		    toMatch = sslen - initialmatchable;
		    /* write out irrelevent prefix, to free buffer */
		    _bfc_writesome(buf, outchan, matchAt, totalwrit, interp);
		    /* line up the buffer perfectly with the believed match */
		    memcpy(buf, buf+matchAt, thistime - matchAt);
		    /* read in just enough more to finish the pattern */
		    if (_bfc_readsome(buf + thistime - matchAt, inchan, 
				      thistimepart, remain, toMatch, 
				      totalread, interp)
			   == TCL_ERROR) {
			return TCL_ERROR;
		    }
		    /* shift back where we look */
		    i -= matchAt;
		    /* adjust how many bits we have to write out */
		    thistime += (thistimepart - matchAt);
		    /* so, did we get the rest of the pattern? */
		    if(strncmp(buf + initialmatchable, 
			       stopstring + initialmatchable, toMatch) == 0) {
			seenstop = 1;
			break;
		    }
		}
		++i;
	    }
	}
	if (seenstop == 0) {
	    _bfc_writesome(buf, outchan, thistime, totalwrit, interp);
	}
    }
    
    sprintf(interp->result, "%d", totalwrit);
    return TCL_OK;
}
