/* 
 * tkPlot.cc 
 *
 *	Display a line in a little widget thing.  Data from a floatRef
 * 
 * dpwe@media.mit.edu 1995aug06
 */

// only do this if we have floatRefs available
#ifdef HAVE_FARRAY

#include <stdlib.h>
#include <string.h>
#include <math.h>	/* for log() proto */
#include <tk.h>
#include <assert.h>

#include <floatArray/floatArray.H>	/* define the floatRef objects frm which we get data */
#include <floatArray/floatArrayTcl.H>	/* name->ptr fns */
//#include "nSurface.h"	/* instead of floatRef.H */
//#include <Otcl.H>	/* definitions for ObjectTcl, packaging for floatRef */

/*
 * A data structure of the following type is kept for each plot
 * widget managed by this file:
 */

typedef struct _plottrace {
    struct _plottrace 	*next;	/* next plot in linked list */
    char*		tag;	/* the tag used to refer to this trace */
    char*		vecname; /* Otcl ref name for the floatRef here */
    floatRef*		fr;	/* reference to the dataset */
    XColor*		color;	/* color to draw this plot in */
    double		base;
    double		len;	/* which part of vector to display */
    double		scale;
    double		offset;	/* conditioning: plot (scale*fr[i])+offset */
    int			doLog;	/* .. take log of fr[i] first */
} PlotTrace;
 
typedef struct {
    Tk_Window	tkwin;		/* Window that embodies the plot.  NULL
				 * means window has been deleted but
				 * widget record hasn't been cleaned up yet. */
    Display	*display;	/* X's token for the window's display. */
    Tcl_Interp	*interp;	/* Interpreter associated with widget. */
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    int 	borderWidth;	/* Width of 3-D border around whole widget. */
//    XColor	*fgColor;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		/* appearance of whole window */
    Tk_Uid orientUid;		/* Orientation for window ("vertical" or
				 * "horizontal"). */
    int vertical;		/* Non-zero means vertical orientation
				 * requested, zero means horizontal. */

    int 	x, y;		/* work area to be used, derived frm win&pad */
    int 	xsize, ysize;

    /*
     * Information used when displaying widget:
     */

    GC gc;			/* Graphics context for copying from
				 * off-screen pixmap onto screen. */
    int doubleBuffer;		/* Non-zero means double-buffer redisplay
				 * with pixmap;  zero means draw straight
				 * onto the display. */
    int updatePending;		/* Non-zero means a call to PlotDisplay
				 * has already been scheduled. */
    Pixmap pm;
    int	   pmw;
    int    pmh;

    /*
     * data source specification
     */
//    char	*vecname;	/* source vecarray tag (can be list) */
//    floatRef	*fv;		/* pointer to the actual data */
//    double	base, len;	/* portion of vec to display */
//    /*
//     * scaling parameters 
//     */
//    double	scale;		/* gray values multiplied by this */
//    double 	offset;		/* added to gray values */
//    int 	doLog;		/* convert raw input data to dB before scale */

    PlotTrace	*traceHead;	/* pointer to first trace record */
    int 	ntraces;	/* total number of traces in list */

} Plot;

/*
 * Information used for argv parsing.
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"#bbbbbb", Tk_Offset(Plot, bgBorder), TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	"white", Tk_Offset(Plot, bgBorder), TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(Plot, borderWidth), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(Plot, doubleBuffer), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(Plot, relief), 0},
    {TK_CONFIG_UID, "-orient", "orient", "Orient",
	"horizontal", Tk_Offset(Plot, orientUid), 0},
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(Plot, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(Plot, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(Plot, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"150", Tk_Offset(Plot, height), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * for "traceplot trace config <tname> ..." commands 
 */

static Tk_ConfigSpec traceConfigSpecs[] = {
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
        "#701040", Tk_Offset(PlotTrace, color), 0 }, 
    {TK_CONFIG_STRING, "-floatref", "floatRef", "FloatRef",
	"", Tk_Offset(PlotTrace, vecname), 0},
    {TK_CONFIG_DOUBLE, "-base", "base", "Base", 
        "0", Tk_Offset(PlotTrace, base), 0}, 
    {TK_CONFIG_DOUBLE, "-len", "len", "Len", 
        "0", Tk_Offset(PlotTrace, len), 0}, 
    {TK_CONFIG_DOUBLE, "-scale", "scale", "Scale", 
        "1", Tk_Offset(PlotTrace, scale), 0}, 
    {TK_CONFIG_DOUBLE, "-offset", "offset", "Offset", 
        "0", Tk_Offset(PlotTrace, offset), 0}, 
    {TK_CONFIG_INT, "-log", "log", "Log", 
	"0", Tk_Offset(PlotTrace, doLog), 0},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};

/*
 * Forward declarations for procedures defined later in this file:
 */

static int		PlotConfigure _ANSI_ARGS_((Tcl_Interp *interp,
			    Plot *plotPtr, int argc, char **argv,
			    int flags));
static void		PlotDestroy _ANSI_ARGS_((char *data));
static void		PlotDisplay _ANSI_ARGS_((ClientData clientData));
static void		PlotEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
static int		PlotWidgetCmd _ANSI_ARGS_((ClientData clientData,
			    Tcl_Interp *, int argc, char **argv));
static int 		PlotReload _ANSI_ARGS_((Tcl_Interp *interp, 
						 Plot *plotPtr));
extern "C" {
int PlotCmd(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
int Plot_Init(Tcl_Interp *interp);
}

int Plot_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "traceplot", PlotCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

static void PlotGetWorkArea(Plot *plotPtr)
{   /* calculate the top left and size of the actual area to draw */
    Tk_Window tkwin = plotPtr->tkwin;

    plotPtr->x = plotPtr->padX+plotPtr->borderWidth;
    plotPtr->y = plotPtr->padY+plotPtr->borderWidth;
    plotPtr->xsize = Tk_Width(tkwin) - 2*(plotPtr->padX+plotPtr->borderWidth);
    plotPtr->ysize = Tk_Height(tkwin) - 2*(plotPtr->padY+plotPtr->borderWidth);
}

/*
 *----------------------------------------------------------------------
 *
 * ptNew --
 *
 *	Create a new PlotTrace linked list element
 *
 * Results:
 *	address of the newly allocated structure
 *
 * Side effects:
 *      sets up the fields of the structure, according to parameters.
 *
 *----------------------------------------------------------------------
 */

PlotTrace *
ptNew(PlotTrace *next, 		/* what to put as the <next> field, to link */
      char *tag)		/* a copy-able string of the tag name to use */
{
    PlotTrace *pt = new PlotTrace;
    pt->next = next;
    pt->tag  = strdup(tag);
    pt->vecname = NULL;
    pt->fr   = NULL;
    pt->color= NULL;
    pt->base = 0;
    pt->len  = 0;
    pt->scale = 1.0;
    pt->offset = 0.0;
    pt->doLog = 0;
    return pt;
}

void
ptFree(PlotTrace *pt)	/* the structure to be destroyed */
{
    if(pt->tag)		free(pt->tag);
    if(pt->vecname)	free(pt->vecname);
    free(pt);
}


/*
 *--------------------------------------------------------------
 *
 * PlotCmd --
 *
 *	This procedure is invoked to process the "traceplot" Tcl
 *	command.  It creates a new "traceplot" widget.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	A new widget is created and configured.
 *
 *--------------------------------------------------------------
 */

int
PlotCmd(ClientData clientData,	/* Main window associated with interpreter. */
	 Tcl_Interp *interp, 	/* Current interpreter. */
	 int argc, 		/* Number of arguments. */
	 char **argv)		/* Argument strings. */
{
    Tk_Window main = (Tk_Window) clientData;
    Plot *plotPtr;
    Tk_Window tkwin;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args:  should be \"",
		argv[0], " pathName ?options?\"", (char *) NULL);
	return TCL_ERROR;
    }

    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], (char *) NULL);
    if (tkwin == NULL) {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Plot");

    /*
     * Allocate and initialize the widget record.
     */

    plotPtr = (Plot *) ckalloc(sizeof(Plot));
    plotPtr->tkwin = tkwin;
    plotPtr->display = Tk_Display(tkwin);
    plotPtr->interp = interp;
    plotPtr->x = 0;
    plotPtr->y = 0;
    plotPtr->xsize = 0;
    plotPtr->ysize = 0;
    plotPtr->padX = 0;
    plotPtr->padY = 0;
    
    plotPtr->borderWidth = 0;
    plotPtr->bgBorder = NULL;
//    plotPtr->fgColor  = NULL;
    plotPtr->relief = TK_RELIEF_FLAT;
    plotPtr->gc = None;
    plotPtr->doubleBuffer = 1;
    plotPtr->updatePending = 0;

//    plotPtr->vecname = NULL;
//    plotPtr->fv      = NULL;

    plotPtr->pm  = None;
    plotPtr->pmh = 0;
    plotPtr->pmw = 0;

//    plotPtr->base = 0;
//    plotPtr->len  = 0;

    plotPtr->traceHead = NULL;
    plotPtr->ntraces   = 0;

    Tk_CreateEventHandler(plotPtr->tkwin, ExposureMask|StructureNotifyMask,
	    PlotEventProc, (ClientData) plotPtr);
    Tcl_CreateCommand(interp, Tk_PathName(plotPtr->tkwin), PlotWidgetCmd,
	    (ClientData) plotPtr, (void (*)(ClientData)) NULL);
    if (PlotConfigure(interp, plotPtr, argc-2, argv+2, 0) != TCL_OK) {
	Tk_DestroyWindow(plotPtr->tkwin);
	return TCL_ERROR;
    }

    interp->result = Tk_PathName(plotPtr->tkwin);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * PlotTagConfigure --
 *
 *	Handle configuration commands for plot tags
 *
 * Results:
 *	Tcl error code
 *
 * Side effects:
 *	Updates the PlotTrace structure with the new config info.
 *	If the plotTracePtrPtr passed in points to a NULL, a new 
 *	PlotTrace structure is created and linked.
 *
 *----------------------------------------------------------------------
 */

int
PlotTagConfigure(Tcl_Interp *interp, 	/* Used for error reporting */
		 Plot *plotPtr,		/* parent structure for this obj */
		 PlotTrace **tracePtrPtr, /* handle for this tag */
		 char *tag, 		/* name of this tag */
		 int argc, 		/* number of valid entries in argv */
		 char **argv)		/* arguments */
{
    int rc = TCL_OK;
    int isnew = 0;
    int flags = TK_CONFIG_ARGV_ONLY;

    if( (*tracePtrPtr)==NULL)  {
	*tracePtrPtr = ptNew(NULL, tag);
	++plotPtr->ntraces;
	isnew = 1;
	flags = 0;	// it's a new block, so set the defaults too
    }
    
    if (Tk_ConfigureWidget(interp, plotPtr->tkwin, traceConfigSpecs,
	    argc, argv, (char *)(*tracePtrPtr), flags) != TCL_OK) {
	if(isnew)  {	// error creating new tag - don't create it!
	    PlotTrace *doomed = *tracePtrPtr;
	    *tracePtrPtr = doomed->next;
	    --plotPtr->ntraces;
	    ptFree(doomed);
	}
	return TCL_ERROR;
    }
    // OK, we successfully applied config to the tag.  Update the other fields
    
    // Make sure a redraw is in order
    rc = PlotReload(interp, plotPtr);
    if(plotPtr->pm != None)  {
	XFreePixmap(Tk_Display(plotPtr->tkwin), plotPtr->pm);
	plotPtr->pm = None;
    }
    if (!plotPtr->updatePending) {
	Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	plotPtr->updatePending = 1;
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * PlotTagCmd --
 *
 *	Processes commands that start <traceplotobj> trace ...
 *	which relate to one of the named traces.
 *
 * Results:
 *	Tcl completion code.
 *
 * Side effects:
 *	Typically configuration of a trace.  Also deletion 
 *	and other tag management functions.
 *
 *----------------------------------------------------------------------
 */

int
PlotTagCmd(Plot* plotPtr, 	/* Information about plot widget. */
	   Tcl_Interp *interp, 	/* Current interpreter. */
	   char *cmd,		/* the object command (original argv[0]) */
	   int argc, 		/* Number of arguments - excluding 'obj tra' */
	   char **argv)		/* Argument strings - starting *after* 'tra' */
{
    int result = TCL_OK;
    int length;
    char c;

    if (argc == 0) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		cmd, " trace <subcmd> ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) plotPtr);
    c = argv[0][0];
    length = strlen(argv[0]);
    // only the 'names' command is *not* followed by a tagname
    if ((c=='n') && (strncmp(argv[0], "names", length)==0))  {
	if(argc != 1)  {
	    Tcl_AppendResult(interp, "wrong #args: should be \"", 
			     cmd, " trace names\"", (char *)NULL);
	    return TCL_ERROR;
	}
	// OK, list the names of all the tags
	PlotTrace *ptp = plotPtr->traceHead;
	while(ptp)  {
	    Tcl_AppendResult(interp, ptp->tag, (ptp->next)?" ":NULL, NULL);
	    ptp = ptp->next;
	}
	return TCL_OK;
    }
    // OK, now must be of form "<obj> trace <subcmd> <tracename> ..."
    if (argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 cmd, " trace <subcmd> <tagname> ?arg ...?\"", NULL);
	return TCL_ERROR;
    }
    char *subcmd = argv[0];
    char *tagname = argv[1];
    argc -= 2;
    argv += 2;
    // quick, find the tag data structure, or point to a NULL if not one
    PlotTrace **ptpp = &(plotPtr->traceHead);
    while(*ptpp)  {
	if(strcmp(tagname, (*ptpp)->tag)==0)
	    break;
	ptpp = &((*ptpp)->next);
    }
    // now *ptpp is either the tag-referred PlotTrace or NULL for unknown tag
    // Process the subcommands
    c = subcmd[0];
    length = strlen(subcmd);
    if ((c == 'c') && (strncmp(subcmd, "create", length) == 0)) {
	if(*ptpp)  {
	    Tcl_AppendResult(interp, "trace \"", tagname, 
			     "\" already exists", NULL);
	    return TCL_ERROR;
	}
	result = PlotTagConfigure(interp, plotPtr, ptpp, tagname, 
				  argc, argv);
    } else if(*ptpp==NULL) {
	// OK, create command is only one valid for a nonexistant tag, 
	// so return an error if the tagname wasn't found already
	goto NoSuchTag;
    } else if ((c == 'c') && (strncmp(subcmd, "configure", length) == 0)) {
	if(argc > 1)  {	// it's a configuration command
	    result = PlotTagConfigure(interp, plotPtr, ptpp, tagname, 
				      argc, argv);
	    if(result != TCL_OK && (*ptpp)==NULL)
		goto NoSuchTag;
	} else if((*ptpp)==NULL)  {	// not allowed to create tag on query
	    goto NoSuchTag;
	} else if (argc == 0) {		// tag definitely exists now
	    result = Tk_ConfigureInfo(interp, plotPtr->tkwin, traceConfigSpecs,
				      (char *) *ptpp, (char *) NULL, 0);
	} else if (argc == 1) {
	    result = Tk_ConfigureInfo(interp, plotPtr->tkwin, traceConfigSpecs,
				      (char *) *ptpp, argv[0], 0);
	}
    } else if ((c == 'c') && (strncmp(subcmd, "cget", length) == 0)) {
	if(argc != 1)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	if((*ptpp)==NULL)  	// not allowed to create tag on query
	    goto NoSuchTag;
	result = Tk_ConfigureValue(interp, plotPtr->tkwin, traceConfigSpecs,
				   (char *) *ptpp, argv[0], 0);
    } else if ((c == 'd') && (strncmp(subcmd, "delete", length)==0))  {
	if((*ptpp)==NULL)
	    goto NoSuchTag;
	PlotTrace *doomed = *ptpp;
	*ptpp = doomed->next;
	--plotPtr->ntraces;
	ptFree(doomed);
	result = PlotReload(interp, plotPtr);
	if(plotPtr->pm != None)  {
	    XFreePixmap(Tk_Display(plotPtr->tkwin), plotPtr->pm);
	    plotPtr->pm = None;
	}
	if (!plotPtr->updatePending) {
	    Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	    plotPtr->updatePending = 1;
	}
    } else {
	Tcl_AppendResult(interp, "bad subcommand \"", subcmd,
		"\":  must be configure, cget, create, delete, names", NULL);
	result = TCL_ERROR;
    }
    return result;

 NoSuchTag:
    Tcl_AppendResult(interp, "traceplot: no such trace \"", tagname, "\"", 
		     NULL);
    return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * PlotWidgetCmd --
 *
 *	This procedure is invoked to process the Tcl command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

static int
PlotWidgetCmd(ClientData clientData, 	/* Information about plot widget. */
	       Tcl_Interp *interp, 	/* Current interpreter. */
	       int argc, 		/* Number of arguments. */
	       char **argv)		/* Argument strings. */
{
    Plot *plotPtr = (Plot *) clientData;
    int result = TCL_OK;
    int length;
    char c;

    if (argc < 2) {
	Tcl_AppendResult(interp, "wrong # args: should be \"",
		argv[0], " option ?arg arg ...?\"", (char *) NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData) plotPtr);
    c = argv[1][0];
    length = strlen(argv[1]);
    if ((c == 'c') && (strncmp(argv[1], "configure", length) == 0)) {
	if (argc == 2) {
	    result = Tk_ConfigureInfo(interp, plotPtr->tkwin, configSpecs,
		    (char *) plotPtr, (char *) NULL, 0);
	} else if (argc == 3) {
	    result = Tk_ConfigureInfo(interp, plotPtr->tkwin, configSpecs,
		    (char *) plotPtr, argv[2], 0);
	} else {
	    result = PlotConfigure(interp, plotPtr, argc-2, argv+2,
		    TK_CONFIG_ARGV_ONLY);
	}
    } else if ((c == 'c') && (strncmp(argv[1], "cget", length) == 0)) {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " cget option\"", (char *)NULL);
	    return TCL_ERROR;
	}
	result = Tk_ConfigureValue(interp, plotPtr->tkwin, configSpecs,
				   (char *) plotPtr, argv[2], 0);
    } else if ((c == 't') && (strncmp(argv[1], "trace", length) == 0)) {
	result = PlotTagCmd(plotPtr, interp, argv[0], argc-2, argv+2); 
    } else if ((c == 'r') && (strncmp(argv[1], "redraw", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " redraw\"", (char *)NULL);
	    return TCL_ERROR;
	}
	if(plotPtr->pm != None)  {
	    XFreePixmap(Tk_Display(plotPtr->tkwin), plotPtr->pm);
	    plotPtr->pm = None;
	}
	if (!plotPtr->updatePending) {
	    Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	    plotPtr->updatePending = 1;
	}
    } else if ((c == 'r') && (strncmp(argv[1], "rescan", length) == 0)) {
	if(argc != 2)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " rescan\"", (char *)NULL);
	    return TCL_ERROR;
	}
	PlotReload(interp, plotPtr);

	if(plotPtr->pm != None)  {
	    XFreePixmap(Tk_Display(plotPtr->tkwin), plotPtr->pm);
	    plotPtr->pm = None;
	}
	if (!plotPtr->updatePending) {
	    Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	    plotPtr->updatePending = 1;
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
		"\":  must be configure, cget, trace, redraw, rescan", (char *) NULL);
	goto error;
    }
    if (!plotPtr->updatePending) {
	Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	plotPtr->updatePending = 1;
    }
    Tk_Release((ClientData) plotPtr);
    return result;

    error:
    Tk_Release((ClientData) plotPtr);
    return TCL_ERROR;
}

floatRef *findFloatRef(char *name)
{   /* load up the plot from a floatArray object */
    int r, c, rs, cs;
    unsigned char *data;
    floatRef *srcFvPtr;
    Tcl_HashEntry *entryPtr;

    //srcFvPtr = (floatRef*)Otcl::otclPtr->obrefToCpp(name,"floatRef");
    srcFvPtr = FRfromName(name);
    if(srcFvPtr == NULL) {  // no floatRef registered to that name
	// If it wasn't a floatRef, maybe it was a floatVec that 
	//  we can cast to a floatRef
	//floatVec *fvp = (floatVec*)Otcl::otclPtr->obrefToCpp(name,"floatVec");
	floatVec *fvp = FVfromName(name);
	if(fvp)  {
	    srcFvPtr = (floatRef*)fvp;
	} else {
	    return NULL;
	}
    }
    return srcFvPtr;
}

static int PlotReload(Tcl_Interp *interp, 	/* Used for error reporting. */
		       Plot *plotPtr)		/* Information about widget. */
{   /* reread the datafile into the internal buffer */
    int rc = TCL_OK;
    PlotTrace *tracePtr = plotPtr->traceHead;
    while(rc==TCL_OK && tracePtr)  {
	if(tracePtr->vecname && strlen(tracePtr->vecname))  {
	    tracePtr->fr = findFloatRef(tracePtr->vecname);
	    if(tracePtr->fr == NULL) {
		Tcl_AppendResult(interp, "no such floatRef: ", 
				 tracePtr->vecname, NULL);
		rc = TCL_ERROR;
	    }
	}
	tracePtr = tracePtr->next;
    }
    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * PlotConfigure --
 *
 *	This procedure is called to process an argv/argc list in
 *	conjunction with the Tk option database to configure (or
 *	reconfigure) a plot widget.
 *
 * Results:
 *	The return value is a standard Tcl result.  If TCL_ERROR is
 *	returned, then interp->result contains an error message.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for plotPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static int
PlotConfigure(Tcl_Interp *interp,	/* Used for error reporting. */
	       Plot *plotPtr, 	/* Information about widget. */
	       int argc, 		/* Number of valid entries in argv. */
	       char **argv, 		/* Arguments. */
	       int flags)		/* Flags to pass to
					 * Tk_ConfigureWidget. */
{
//    int needNewData = (plotPtr->data==NULL)?1:0;
    int needNewData = 1; 	/* if there are args, rescan data */
    int rc = TCL_OK;

    if (Tk_ConfigureWidget(interp, plotPtr->tkwin, configSpecs,
	    argc, argv, (char *) plotPtr, flags) != TCL_OK) {
	return TCL_ERROR;
    }
    /* Convert orientation string to actual direction */
    int length = strlen(plotPtr->orientUid);
    if (strncmp(plotPtr->orientUid, "vertical", length) == 0) {
	plotPtr->vertical = 1;
    } else if (strncmp(plotPtr->orientUid, "horizontal", length) == 0) {
	plotPtr->vertical = 0;
    } else {
	Tcl_AppendResult(interp, "bad orientation \"", plotPtr->orientUid,
		"\": must be vertical or horizontal", (char *) NULL);
	return TCL_ERROR;
    }

    /*
     * Set the background for the window and create a graphics context
     * for use during redisplay.
     */

    Tk_SetWindowBackground(plotPtr->tkwin,
	    Tk_3DBorderColor(plotPtr->bgBorder)->pixel);
    if ((plotPtr->gc == None) && (plotPtr->doubleBuffer)) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	plotPtr->gc = Tk_GetGC(plotPtr->tkwin,
		GCFunction|GCGraphicsExposures, &gcValues);
    }

    /*
     * Register the desired geometry for the window.  Then arrange for
     * the window to be redisplayed.
     */

    Tk_GeometryRequest(plotPtr->tkwin, plotPtr->width, plotPtr->height);
    Tk_SetInternalBorder(plotPtr->tkwin, plotPtr->borderWidth);

    if(needNewData)  {
	rc = PlotReload(interp, plotPtr);
	// force redraw by disposing of our existing pixmap
	if(rc==TCL_OK && plotPtr->pm != None)  {
	    XFreePixmap(Tk_Display(plotPtr->tkwin), plotPtr->pm);
	    plotPtr->pm = None;
	}
    }

    if (!plotPtr->updatePending) {
	Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	plotPtr->updatePending = 1;
    }

    return rc;
}

/*
 *--------------------------------------------------------------
 *
 * PlotEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on plots.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
PlotEventProc(ClientData clientData,	/* Information about window. */
	       XEvent *eventPtr)	/* Information about event. */
{
    Plot *plotPtr = (Plot *) clientData;

    if (eventPtr->type == Expose) {
	if (!plotPtr->updatePending) {
	    Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	    plotPtr->updatePending = 1;
	}
    } else if (eventPtr->type == ConfigureNotify) {
	if (!plotPtr->updatePending) {
	    Tk_DoWhenIdle(PlotDisplay, (ClientData) plotPtr);
	    plotPtr->updatePending = 1;
	}
    } else if (eventPtr->type == DestroyNotify) {
	Tcl_DeleteCommand(plotPtr->interp, Tk_PathName(plotPtr->tkwin));
	plotPtr->tkwin = NULL;
	if (plotPtr->updatePending) {
	    Tk_CancelIdleCall(PlotDisplay, (ClientData) plotPtr);
	}
	Tcl_EventuallyFree((ClientData) plotPtr, (Tcl_FreeProc *)PlotDestroy);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * PlotDrawTrace --
 *
 *	Draw the contents of a single trace onto the plot panel
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Performs graphics commands upon the specified drawable to 
 *	display a trace reflecting the data in the passed floatRef
 *
 *----------------------------------------------------------------------
 */

void
PlotDrawTrace(Plot *plotPtr,		/* the object's parent structure */
	      PlotTrace *tracePtr, 	/* the particular trace to draw */
	      Display *disp, 
	      Drawable d)		/* where to draw it */
{
    /* draw the line */
    double	scale   = tracePtr->scale;
    int		d0, dscale, dd=0;
    int		xo, yo;
    int		xd = 0;
    int		yd = 0;
    int		x = plotPtr->x;
    int		y = plotPtr->y;
    int		w = plotPtr->xsize;
    int		h = plotPtr->ysize;
    int i;
    int nxs, xssize = 100;
    XSegment *xsgs, *thisSeg;
    XRectangle      cliprect;
    int		flen  = tracePtr->fr->Size();
    int		vbase = tracePtr->base;
    int		vlen  = tracePtr->len;
    double	da;
    int 	notfirsttime = 0;
	
    /* vbase+vlen specify which part of the floatRef to actually 
       show.  If vlen is zero, it means go to the end of the array */
    if(vlen <= 0)
	vlen = flen - max(0,vbase);

    if(vlen < 2)
	vlen = 2;	/* minimum */

    cliprect.x = x; cliprect.y = y; cliprect.height = h; cliprect.width =w;
    XSetClipRectangles(disp, plotPtr->gc, 0, 0, &cliprect, 1, YXSorted);

    if(plotPtr->vertical)  {
	dscale = -w/2;
	d0     = x+w/2 + (int)(tracePtr->offset*dscale);
    } else {	// horizontal
	dscale = -h/2;
	d0     = y+h/2 + (int)(tracePtr->offset*dscale);
    }
    XSetForeground(disp, plotPtr->gc, tracePtr->color->pixel);
    
    xsgs = (XSegment *)ckalloc(xssize * sizeof(XSegment));
    thisSeg = xsgs;
    nxs = 0;
    for(i = vbase; i < (vbase+vlen); ++i)  {
	if(i>=0 && i<flen)  { /* is a point within range */
	    yo = yd;  xo = xd;
	    da = tracePtr->fr->El(i);
	    if(tracePtr->doLog)  {
		dd = int(d0+dscale*scale*log(max(da, 0.000001)));
	    } else{
		dd = int(d0+dscale*scale*da);
	    }
	    if(plotPtr->vertical)  {
		xd = dd;	// len-1 because that's the # of segs
		yd = y+h-1-(((i-vbase)*h)/(vlen-1)); 
	    } else {				 // horizontal
		yd = dd;	// len-1 because that's the # of segs
		xd = x+(((i-vbase)*w)/(vlen-1)); 
	    }
	    if(notfirsttime)  {
		thisSeg->x1 = xo;	thisSeg->y1 = yo;
		thisSeg->x2 = xd;	thisSeg++->y2 = yd;
		if(++nxs == xssize)  {
		    XDrawSegments(disp, d, plotPtr->gc, xsgs, nxs);
		    thisSeg = xsgs;
		    nxs = 0;
		}
	    } else {
		notfirsttime = 1;
	    }
	}
    }
    if(nxs > 0)  {
	XDrawSegments(disp, d, plotPtr->gc, xsgs, nxs);
    }
    /* Disable clipping now we're done */
    XSetClipMask(disp, plotPtr->gc, None);
	
    ckfree((char *)xsgs);
}

/*
 *--------------------------------------------------------------
 *
 * PlotDisplay --
 *
 *	This procedure redraws the contents of a plot window.
 *	It is invoked as a do-when-idle handler, so it only runs
 *	when there's nothing else for the application to do.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
PlotDisplay(ClientData clientData)	/* Information about window. */
{
    Plot *plotPtr = (Plot *) clientData;
    Tk_Window tkwin = plotPtr->tkwin;
    Drawable d;
    Display  *disp = Tk_Display(tkwin);
    int w,h,v;

    plotPtr->updatePending = 0;
    if (!Tk_IsMapped(tkwin)) {
	return;
    }

    /* Create a pixmap for double-buffering, if necessary. */
    w = Tk_Width(tkwin); h = Tk_Height(tkwin);
    if (plotPtr->doubleBuffer) {
	if(plotPtr->pm == None || plotPtr->pmw != w || plotPtr->pmh != h) {
	    if(plotPtr->pm != None)
		XFreePixmap(Tk_Display(tkwin), plotPtr->pm);
	    plotPtr->pm = XCreatePixmap(disp, Tk_WindowId(tkwin),
					 w, h, 
				     DefaultDepthOfScreen(Tk_Screen(tkwin)));
	    plotPtr->pmw = w;
	    plotPtr->pmh = h;
	    /* Redraw the widget's background and border only once per pm. */
	    Tk_Fill3DRectangle(tkwin, plotPtr->pm, 
			       plotPtr->bgBorder,
			       0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
			       plotPtr->borderWidth, plotPtr->relief);

	}
	d = plotPtr->pm;
    } else {
	d = Tk_WindowId(tkwin);
	/* Redraw the widget's background and border. */
	Tk_Fill3DRectangle(tkwin, d, plotPtr->bgBorder,
			   0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
			   plotPtr->borderWidth, plotPtr->relief);
    }

    /* Display the plot. */
    /* first, set up work area ptr (->x, ->xsize etc) */
    PlotGetWorkArea(plotPtr);

    int ndrawn = 0;
    PlotTrace *tracePtr = plotPtr->traceHead;
    while(tracePtr)  {
	floatRef *fr = tracePtr->fr;
	if(fr) {
	    PlotDrawTrace(plotPtr, tracePtr, disp, d);
	    ++ndrawn;
	}
	tracePtr = tracePtr->next;
    }
    if(ndrawn == 0)  {	// nothing was drawn
	Tk_Fill3DRectangle(tkwin, d, plotPtr->bgBorder,
			   plotPtr->x, plotPtr->y, 
			   plotPtr->xsize, plotPtr->ysize,
			   plotPtr->borderWidth, TK_RELIEF_SUNKEN);
    }
    /* If double-buffered, copy to the screen */
    if (plotPtr->doubleBuffer) {
	XCopyArea(disp, plotPtr->pm, Tk_WindowId(tkwin), 
		  plotPtr->gc,
		  0, 0, Tk_Width(tkwin), Tk_Height(tkwin), 0, 0);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * PlotDestroy --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a plot at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the plot is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
PlotDestroy(char *data) /* Info about plot widget. */
{
    Plot *plotPtr = (Plot *)data;
    int i;
    
    Tk_FreeOptions(configSpecs, (char *) plotPtr, plotPtr->display, 0);
    if (plotPtr->gc != None) {
	Tk_FreeGC(plotPtr->display, plotPtr->gc);
    }
    /*  if(plotPtr->data)  {
	ckfree((char *)plotPtr->data);
	plotPtr->data = NULL;
	}
	*/
    ckfree((char *) plotPtr);
}

#endif /* HAVE_FARRAY */
